from typing import Any
from typing import Dict
from typing import Generic
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import TypeVar

import attr

from avalon.common.errors import SwitchError
from avalon.datagen.godot_env.observations import AvalonObservation

# Mapping of feature name to (data_type, shape).
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.constants import avalon_task_to_int
from avalon.datagen.world_creation.world_generator import GenerateAvalonWorldParams
from avalon.datagen.world_creation.world_generator import GeneratedAvalonWorldParams
from avalon.datagen.world_creation.world_generator import GeneratedWorldParamsType

ObsType = TypeVar("ObsType")


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class GoalProgressResult:
    reward: float
    is_done: bool
    log: Dict[str, Any]
    world_path: Optional[str] = None


class GoalEvaluator(Generic[ObsType, GeneratedWorldParamsType]):
    def calculate_next_is_done_and_reward(self, observation: ObsType) -> Tuple[bool, float]:
        result = self.calculate_goal_progress(observation)
        return result.is_done, result.reward

    def calculate_goal_progress(self, observation: ObsType) -> GoalProgressResult:
        raise NotImplementedError()

    def reset(self, observation: ObsType, world_params: Optional[GeneratedWorldParamsType] = None) -> None:
        raise NotImplementedError()


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class NullGoalEvaluator(GoalEvaluator[AvalonObservation, GeneratedAvalonWorldParams]):
    def calculate_goal_progress(self, observation: AvalonObservation) -> GoalProgressResult:
        return GoalProgressResult(reward=0, is_done=False, log={})

    def reset(self, observation: AvalonObservation, world_params: Optional[GeneratedAvalonWorldParams] = None) -> None:
        pass


FRAMES_PER_MINUTE = 600


@attr.s(auto_attribs=True, collect_by_mro=True)
class AvalonGoalEvaluator(GoalEvaluator[AvalonObservation, GeneratedAvalonWorldParams]):
    score_cost_per_frame: float = 1e-4
    frame_limit: int = 9000

    def calculate_goal_progress(self, observation: AvalonObservation) -> GoalProgressResult:
        remaining_frames = self.frame_limit - observation.frame_id.item() - 1
        if not self.is_all_food_eaten:  # type: ignore[has-type]
            self.score = max(
                0,
                observation.hit_points.item() + remaining_frames * self.score_cost_per_frame - self.initial_hit_points,
            )
        self.is_all_food_eaten = bool(observation.is_food_present_in_world.item() < 0.5)

        is_done = observation.is_done.item() > 0

        truncated = False
        if not is_done and remaining_frames == 0:
            is_done = True
            truncated = True

        assert self.world_params is not None
        return GoalProgressResult(
            is_done=is_done,
            reward=self.get_reward(observation),
            log={
                "success": 1 if self.is_all_food_eaten else 0,
                "score": self.score,
                "remaining_frames": remaining_frames,
                "frame_id": observation.frame_id.item(),
                "difficulty": self.world_params.difficulty,
                # make this an int to work with our rollout worker
                "task": avalon_task_to_int[self.world_params.task.value],
                "episode_id": observation.episode_id.item(),
                "world_index": self.world_params.index,
                "hit_points": observation.hit_points.item(),
                "TimeLimit.truncated": truncated,
            },
            world_path=self.world_params.output,
        )

    def get_reward(self, observation: AvalonObservation) -> float:
        return observation.reward.item() - self.score_cost_per_frame

    def get_level_frame_limit(self, world_params: GenerateAvalonWorldParams) -> int:
        if world_params.task in (AvalonTask.FIND, AvalonTask.GATHER, AvalonTask.NAVIGATE):
            return 15 * FRAMES_PER_MINUTE
        elif world_params.task in (AvalonTask.SURVIVE, AvalonTask.STACK, AvalonTask.CARRY, AvalonTask.EXPLORE):
            return 10 * FRAMES_PER_MINUTE
        else:
            return 5 * FRAMES_PER_MINUTE

    def reset(self, observation: AvalonObservation, world_params: Optional[GenerateAvalonWorldParams] = None) -> None:
        assert world_params is not None
        self.world_params = world_params
        self.initial_hit_points = observation.hit_points.item()
        self.frame_limit = self.get_level_frame_limit(world_params)
        self.score = self.frame_limit * self.score_cost_per_frame
        self.is_all_food_eaten = bool(observation.is_food_present_in_world.item() < 0.5)


# These are calculated such that 98% of humans succeed after applying the dynamic frame limit
TRAINING_FRAME_LIMITS_AT_ZERO_DIFFICULTY = {
    AvalonTask.EAT: 260,
    AvalonTask.FIGHT: 220,
    AvalonTask.PUSH: 700,
    AvalonTask.DESCEND: 340,
    AvalonTask.STACK: 710,
    AvalonTask.CLIMB: 250,
    AvalonTask.SCRAMBLE: 190,
    AvalonTask.THROW: 600,
    AvalonTask.GATHER: 3300,
    AvalonTask.CARRY: 610,
    AvalonTask.HUNT: 420,
    AvalonTask.FIND: 1180,
    AvalonTask.EXPLORE: 1340,
    AvalonTask.MOVE: 360,
    AvalonTask.BRIDGE: 630,
    AvalonTask.SURVIVE: 3780,
    AvalonTask.OPEN: 170,
    AvalonTask.NAVIGATE: 1220,
    AvalonTask.AVOID: 240,
    AvalonTask.JUMP: 310,
}


@attr.s(auto_attribs=True, collect_by_mro=True)
class TrainingAvalonGoalEvaluator(AvalonGoalEvaluator):
    energy_cost_coefficient: float = 1e-9
    body_ke_coefficient: float = 0.0
    body_pe_coefficient: float = 1.0
    head_pe_coefficient: float = 1.0
    hand_ke_coefficient: float = 0.0
    hand_pe_coefficient: float = 1.0
    head_roll_coefficient: float = 0.0
    head_pitch_coefficient: float = 0.0
    energy_cost_aggregator: Literal["sum", "max"] = "sum"

    def get_level_frame_limit(self, world_params: GenerateAvalonWorldParams) -> int:
        super_frame_limit = super().get_level_frame_limit(world_params)
        dynamic_frame_limit = int(
            TRAINING_FRAME_LIMITS_AT_ZERO_DIFFICULTY[world_params.task] * 10**world_params.difficulty
        )
        return min(super_frame_limit, dynamic_frame_limit)

    def get_reward(self, observation: AvalonObservation) -> float:
        energy_cost = (
            self.energy_cost_coefficient
            * (
                self.head_pe_coefficient * observation.physical_head_potential_energy_expenditure.item()
                + self.body_ke_coefficient * observation.physical_body_kinetic_energy_expenditure.item()
                + self.body_pe_coefficient * observation.physical_body_potential_energy_expenditure.item()
                + self.hand_ke_coefficient * observation.physical_left_hand_kinetic_energy_expenditure.item()
                + self.hand_pe_coefficient * observation.physical_left_hand_potential_energy_expenditure.item()
                + self.hand_ke_coefficient * observation.physical_right_hand_kinetic_energy_expenditure.item()
                + self.hand_pe_coefficient * observation.physical_right_hand_potential_energy_expenditure.item()
            )
            + self.head_roll_coefficient * observation.physical_head_rotation[2].item()
            + self.head_pitch_coefficient * max(abs(observation.physical_head_rotation[0].item()) - 45.0, 0.0)
        )
        if self.energy_cost_aggregator == "sum":
            energy_cost += self.score_cost_per_frame
        elif self.energy_cost_aggregator == "max":
            energy_cost = max(self.score_cost_per_frame, energy_cost)
        else:
            raise SwitchError(f"Invalid energy_cost_aggregator: {self.energy_cost_aggregator}")
        return float(observation.reward.item() - energy_cost)



import attr
import godot_parser
from godot_parser import Node as GDNode

from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.types import GodotScene


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ColoredSphere(Tool):
    resource_file: str = "res://items/sphere_marker.tscn"
    color: str = "#0000FF"
    scale: float = 2.0

    def get_node(self, scene: GodotScene) -> GDNode:
        node = super().get_node(scene)
        node.properties["color"] = self.color
        node.properties["scale"] = godot_parser.Vector3(self.scale, self.scale, self.scale)
        return node



import attr

from avalon.datagen.world_creation.types import SceneryBorderMode


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class SceneryConfig:
    # the file that defines the object
    resource_file: str
    # the biome into which to spawn the items
    biome_id: int
    # number of items per square meter (at density=1.0)
    density: float = 0.2
    # how the density varies in the distances between 0 and border_distance
    border_mode: SceneryBorderMode = SceneryBorderMode.LINEAR
    # over what border range to smooth between 0.0 and 1.0
    border_distance: float = 5.0
    # scale of the objects will be +/- this fraction on all axes
    correlated_scale_range: float = 0.4
    # scale of the objects will be +/- this fraction on EACH axis
    # ie, will squish and stretch the object
    skewed_scale_range: float = 0.2
    # if set to True, will align the item with the normal of the surface
    is_oriented_to_surface: bool = False



import attr
from loguru import logger

from avalon.agent.common.params import DmcEnvironmentParams
from avalon.agent.common.parse_args import parse_args
from avalon.agent.common.util import setup_new_process
from avalon.agent.dreamer.params import DreamerParams
from avalon.agent.train_dreamer_dmc import DreamerTrainer
from avalon.common.log_utils import configure_remote_logger

"""
- note: the time limit appears to be applied in agent env steps, not (action-repeated) env_steps.
- the exact model params aren't entirely clear for the atari continuous latents.
  - maksis ran: hidden 600, deter 600, stoch 32
    - does this line up with the paper curves?
  - and the dreamerv1 config default was: hidden 400 (?), deter 200, stoch 30
  - hidden -> just for mlps. deter is the hidden,state size of GRU.
"""


@attr.s(auto_attribs=True, frozen=True)
class DreamerAtariParams(DreamerParams):
    name: str = "torch"
    total_env_steps: int = 50_000_000
    pcont: bool = True
    prioritize_ends: bool = True
    worker_managers: int = 1
    num_workers: int = 1
    env_params: DmcEnvironmentParams = DmcEnvironmentParams(
        suite="atari", task="atlantis", action_repeat=4, time_limit=27000
    )
    actor_grad: str = "reinforce"
    free_nats: float = 0
    policy_entropy_scale: float = 1e-3
    model_lr: float = 2e-4
    actor_lr: float = 4e-5
    value_lr: float = 1e-4
    rssm_hidden_size: int = 600
    deter_size: int = 600
    discount: float = 0.999
    kl_loss_scale: float = 0.1
    pcont_loss_scale: float = 5
    prefill_steps: int = 50000
    pretrain_steps: int = 1
    freeze_actor_steps: int = 0
    train_every: int = 16
    train_gpu: int = 0
    inference_gpus: tuple[int, ...] = (0,)
    log_freq_hist: int = 2000
    log_freq_scalar: int = 25
    log_freq_media: int = 1000
    checkpoint_every: int = 10_000
    num_dataloader_workers: int = 1
    log_rollout_metrics_every: int = 1
    prefill_eps_per_dataloader: int = 1


def run(params: DreamerAtariParams) -> None:
    # params = attr.evolve(params, name=f"{params.name}_{params.env_params.suite}_{params.env_params.task}")
    trainer = DreamerTrainer(params)
    try:
        trainer.train()
    finally:
        trainer.shutdown()


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    try:
        default_params = DreamerAtariParams()
        default_params = parse_args(default_params)
        run(default_params)
    except Exception as e:
        logger.exception(e)
        raise



import attr

from avalon.datagen.world_creation.entities.tools.stone import Stone


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Boulder(Stone):
    mass: float = 60.0
    resource_file: str = "res://items/stone.tscn"
    script_file: str = "res://items/boulder.gd"
    size: float = 1.4



from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

import attr
import numpy as np
from loguru import logger
from matplotlib import pyplot as plt
from nptyping import assert_isinstance
from numpy.typing import NDArray
from scipy import ndimage
from scipy import stats
from scipy.interpolate import RectBivariateSpline
from scipy.interpolate import RegularGridInterpolator
from scipy.ndimage import convolve
from scipy.ndimage import gaussian_filter
from scipy.ndimage import generate_binary_structure
from scipy.spatial import KDTree
from skimage import morphology
from skimage.morphology import flood_fill

from avalon.common.errors import SwitchError
from avalon.contrib.serialization import Serializable
from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.world_creation.configs.biome import BiomeConfig
from avalon.datagen.world_creation.configs.flora import FloraConfig
from avalon.datagen.world_creation.configs.world import WorldConfig
from avalon.datagen.world_creation.constants import WATER_LINE
from avalon.datagen.world_creation.debug_plots import IS_DEBUG_VIS
from avalon.datagen.world_creation.debug_plots import plot_points
from avalon.datagen.world_creation.debug_plots import plot_value_grid
from avalon.datagen.world_creation.noise import perlin
from avalon.datagen.world_creation.region import EdgedRegion
from avalon.datagen.world_creation.region import Region
from avalon.datagen.world_creation.types import HeightMode
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import MapFloatNP
from avalon.datagen.world_creation.types import Point2DListNP
from avalon.datagen.world_creation.types import Point2DNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.types import SupportsIndexTuple


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class SpecialBiomes:
    beach_mask: MapBoolNP
    swamp_mask: MapBoolNP
    fresh_water_mask: MapBoolNP


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class VisibilityCalculator:
    cells_per_meter: float
    height_interpolator: RegularGridInterpolator

    def is_visible_from(self, a: Point3DNP, b: Point3DNP, is_plotted: bool = False) -> bool:
        if np.isclose(a, b).all():
            return True
        a2 = np.array([a[0], a[2]])
        b2 = np.array([b[0], b[2]])
        total_distance_2d = np.linalg.norm((b2 - a2))
        # looking at every other cell seems relatively sasfe, the heights are usually not too variable
        point_count = max(round((self.cells_per_meter * total_distance_2d) * 0.5), 5)
        points = np.linspace(a2, b2, point_count)
        distances = np.linspace(0.0, total_distance_2d, point_count)
        # it's idiotic that these need to be reversed, whatever
        swapped_points = points.copy()
        swapped_points[:, [1, 0]] = swapped_points[:, [0, 1]]
        heights = self.height_interpolator(swapped_points)
        a_height = a[1]
        b_height = b[1]
        # always go from the lowest point to the highest, makes it easier to think about
        if b_height < a_height:
            heights = np.flip(heights)
            b_height, a_height = a_height, b_height
        height_deltas = heights - a_height
        end_height = b_height - a_height
        if is_plotted:
            plt.plot(distances, height_deltas)
            plt.show()
        final_height_dist_ratio = end_height / distances[-1]
        height_dist_ratios = height_deltas[1:-1] / distances[1:-1]
        result = not np.any(height_dist_ratios > final_height_dist_ratio)
        return result


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HeightMap:
    region: Region
    cells_per_meter: float
    X: MapFloatNP
    Y: MapFloatNP
    Z: MapFloatNP
    _simplicity_warnings: List[str]

    @staticmethod
    def create(region: Region, cells_per_meter: float) -> "HeightMap":
        x, y = region.get_linspace(cells_per_meter)
        X, Y = np.meshgrid(x, y)
        Z = 0.0 * X
        return HeightMap(region, cells_per_meter, X, Y, Z, [])

    def copy(self) -> "HeightMap":
        return HeightMap(
            self.region,
            self.cells_per_meter,
            self.X.copy(),
            self.Y.copy(),
            self.Z.copy(),
            [x for x in self._simplicity_warnings],
        )

    def freeze(self) -> "FrozenHeightMap":
        X = self.X.copy()
        X.setflags(write=False)
        Y = self.Y.copy()
        Y.setflags(write=False)
        Z = self.Z.copy()
        Z.setflags(write=False)
        return FrozenHeightMap(
            self.region,
            self.cells_per_meter,
            X,
            Y,
            Z,
            tuple(self._simplicity_warnings),
        )

    def plot(self):
        plt.figure(figsize=(20, 20))
        ax = plt.subplot(projection="3d")
        meshcount = 129
        gridcount = len(self.X[0, :])
        if gridcount > meshcount:
            logger.warning(f"WARNING: Mesh is {gridcount / meshcount}x finer than can be displayed")
        ax.plot_wireframe(self.X, self.Y, self.Z, color="k", rcount=meshcount, ccount=meshcount)
        rough_meters = (self.region.x.size + self.region.z.size) / 2.0
        ax.set_zlim(-rough_meters / 2.0, rough_meters / 2.0)
        ax.set_axis_off()
        # logger.debug(ax.elev)
        # logger.debug(ax.azim)
        plt.show()
        return self

    def add_noise(self, scale: float, rand: np.random.Generator) -> None:
        self.Z += rand.uniform(-scale / 2.0, scale / 2.0, self.Z.shape)

    def add_center_biased_noise(
        self, mountain_center: Tuple[float, float], scale: float, mountain_radius: float, rand: np.random.Generator
    ) -> None:
        noise = rand.normal(scale, scale / 3.0, self.Z.shape)
        max_dist = mountain_radius * self.region.x.size
        dist_sq_to_mountain = self.get_dist_sq_to(np.array(mountain_center))
        weight = 1.0 - (dist_sq_to_mountain / (max_dist**2))
        self.Z = self.Z + np.clip(weight, 0.0, 1.0) * noise

    def get_land_mask(self) -> MapBoolNP:
        return self.Z > 0.0

    def apply_height_mask(self, mask: np.ndarray, height: float, height_mode: HeightMode):
        # basically, sinking one part of the world is equivalent to raising the rest, EXCEPT at the water
        # if there is a negative height difference that would put any part of the region underwater
        # we have to raise the whole rest of the landmass, and mark the coast as unclimbable (if not already done)
        # otherwise, you could simply walk around the island to avoid having to cross chasms

        # figure out everywhere the mask applies to land:
        land_mask = self.get_land_mask()
        if height_mode == HeightMode.RELATIVE:
            if height < 0:
                # plot_value_grid(self.Z, "height when applying")
                # plot_value_grid(mask, "mask")
                effect_mask = np.logical_and(land_mask, mask > 0)
                if not np.any(effect_mask):
                    # plot_value_grid(mask, "doesn't matter mask")
                    return
                min_height = self.Z[effect_mask].min()
                height_delta = min_height + height
                if height_delta < 0:
                    # shift the world up
                    self.Z[land_mask] += -height_delta
                    assert (
                        False
                    ), "Should no longer be able to get here--use the world begin_height_obstacles and end_height_obstacles methods instead"
            self.Z += height * mask
        elif height_mode == HeightMode.ABSOLUTE:
            assert height > 0
            self.Z[np.logical_and(land_mask, mask >= 1.0)] = height
            interpolation_mask = np.logical_and(land_mask, np.logical_and(mask < 1.0, mask > 0.0))
            self.Z[interpolation_mask] = (self.Z[interpolation_mask] - height) * (
                1.0 - mask[interpolation_mask]
            ) + height
        else:
            raise SwitchError(f"Unknown HeightMode: {height_mode}")

    def upsample(self, factor: float = 2.0) -> "HeightMap":
        x, y = self.region.get_linspace(self.cells_per_meter)
        interp_spline = RectBivariateSpline(x, y, self.Z)

        x2 = np.linspace(
            self.region.x.min_ge, self.region.x.max_lt, round(self.region.x.size * self.cells_per_meter * factor) + 1
        )
        y2 = np.linspace(
            self.region.z.min_ge, self.region.z.max_lt, round(self.region.z.size * self.cells_per_meter * factor) + 1
        )

        X2, Y2 = np.meshgrid(x2, y2)
        Z2 = interp_spline(x2, y2)

        return HeightMap(self.region, self.cells_per_meter * factor, X2, Y2, Z2, self._simplicity_warnings)

    def get_rough_height_at_point(self, point: Point2DNP) -> float:
        assert_isinstance(point, Point2DNP)
        return self.Z[self.point_to_index(point)]  # type: ignore

    def get_2d_points(self):
        return np.stack([self.X, self.Y], axis=2)

    def get_3d_points(self):
        return np.stack([self.X, self.Z, self.Y], axis=2)

    def get_lineseg_distances(self, start: Point2DNP, end: Point2DNP):
        a = start.reshape((1, 2))
        b = end.reshape((1, 2))
        points_2d = self.get_2d_points()
        line_seg_distances = lineseg_dists(points_2d.reshape((-1, 2)), a, b).reshape(points_2d.shape[:2])
        return line_seg_distances

    def get_heights(self, points: NDArray[np.floating]) -> NDArray[np.floating]:
        assert_isinstance(points, Point2DListNP)
        points_2d = points.copy()
        # y before x internally
        points_2d[:, [1, 0]] = points_2d[:, [0, 1]]
        grid_x, grid_y = self.region.get_linspace(self.cells_per_meter)
        interp_spline = RegularGridInterpolator((grid_y, grid_x), self.Z)
        return cast(NDArray[np.floating], interp_spline(points_2d))

    def restrict_points_to_region(self, points: Point2DListNP) -> Point2DListNP:
        assert_isinstance(points, Point2DListNP)
        mask = np.logical_and(
            np.logical_and(points[:, 0] >= self.region.x.min_ge, points[:, 0] <= self.region.x.max_lt),
            np.logical_and(points[:, 1] >= self.region.z.min_ge, points[:, 1] <= self.region.z.max_lt),
        )
        return cast(Point2DListNP, points[mask])

    def points_to_indices(self, points: Point2DListNP) -> Tuple[np.ndarray, np.ndarray]:
        assert_isinstance(points, Point2DListNP)
        max_x_index = self.Z.shape[1] - 1
        x = np.rint(np.interp(points[:, 0], [self.region.x.min_ge, self.region.x.max_lt], [0, max_x_index])).astype(
            np.int32
        )
        max_y_index = self.Z.shape[0] - 1
        y = np.rint(np.interp(points[:, 1], [self.region.z.min_ge, self.region.z.max_lt], [0, max_y_index])).astype(
            np.int32
        )
        return x, y

    def get_dist_sq_to(self, point: Point2DNP):
        assert_isinstance(point, Point2DNP)
        locations = self.get_2d_points()
        dist_sq = (locations[:, :, 0] - point[0]) ** 2 + (locations[:, :, 1] - point[1]) ** 2
        return dist_sq

    def to_mutable(self) -> "HeightMap":
        return self.copy()

    def radial_flatten(self, point: Point2DNP, radius: float, extra_mask: Optional[MapBoolNP] = None):
        # height = only(self.get_heights(np.array([point])))
        height = self.get_rough_height_at_point(point)
        assert height > WATER_LINE, "Cannot flatten points in the ocean"
        dist_sq = self.get_dist_sq_to(point)

        mask = dist_sq < radius * radius
        if extra_mask is not None:
            mask = np.logical_and(extra_mask, mask)
        mixing = np.clip(((radius - np.sqrt(dist_sq[mask])) / radius) * 2, 0, 1)
        self.Z[mask] = self.Z[mask] * (1.0 - mixing) + height * mixing

        # plot_value_grid(self.Z)

        return dist_sq

    def get_random_land_point(self, rand: np.random.Generator, ideal_shore_dist: float) -> Point2DNP:
        """:returns: a random point that is guaranteed to be on the land"""
        land_mask = self.get_land_mask()
        if ideal_shore_dist > 0.0:
            original_land_mask = land_mask
            shore_cell_dist = int(self.cells_per_meter * ideal_shore_dist) + 1
            land_mask = np.logical_not(
                morphology.dilation(np.logical_not(land_mask), morphology.disk(shore_cell_dist))
            )
            if not np.any(land_mask):
                land_mask = original_land_mask
        possible_points = np.argwhere(land_mask)
        indices = tuple(rand.choice(possible_points))
        x = self.X[indices]
        y = self.Y[indices]
        return np.array([x, y])

    def get_island(self, point: Point2DNP) -> Tuple[MapBoolNP, MapBoolNP]:
        assert_isinstance(point, Point2DNP)
        index = self.point_to_index(point)
        mask = self.get_land_mask().astype(int)
        flood_fill(mask, index, 2, in_place=True)
        island_mask = mask == 2

        water_mask = np.logical_not(self.get_land_mask())
        ocean_mask = water_mask.copy().astype(int)
        flood_fill(ocean_mask, (0, 0), 2, in_place=True)
        ocean_mask = ocean_mask == 2
        fresh_water_mask = np.logical_and(water_mask, np.logical_not(ocean_mask))

        island_mask_without_fresh_water = island_mask.astype(int).copy()
        island_mask_without_fresh_water[fresh_water_mask] = 2
        island_mask_without_fresh_water[ocean_mask] = -100
        flood_fill(island_mask_without_fresh_water, index, 3, tolerance=3, in_place=True)
        island_mask_without_fresh_water = island_mask_without_fresh_water == 3
        # plot_value_grid(island_mask_without_fresh_water, "Island without fresh water")

        return island_mask, island_mask_without_fresh_water

    def generate_visibility_calculator(self) -> VisibilityCalculator:
        """Optimization. Here so we dont have to make this mesh grid multiple times if checking visibility many times"""
        grid_x, grid_y = self.region.get_linspace(self.cells_per_meter)
        height_interpolator = RegularGridInterpolator((grid_y, grid_x), self.Z)
        return VisibilityCalculator(self.cells_per_meter, height_interpolator)

    # TODO: is gross that this, unlike everythign else, accepts only 3d points...
    def get_normals(self, points: np.ndarray) -> np.ndarray:
        delta = 0.5
        points_2d = np.stack([points[:, 2], points[:, 0]], axis=1)

        grid_x, grid_y = self.region.get_linspace(self.cells_per_meter)
        interp_spline = RegularGridInterpolator((grid_y, grid_x), self.Z)
        heights_at_points = interp_spline(points_2d)
        heights_right = interp_spline(points_2d + np.array([0.0, delta]))
        heights_up = interp_spline(points_2d + np.array([delta, 0.0]))
        points_3d = np.stack([points[:, 0], heights_at_points, points[:, 1]], axis=1)
        points_right = np.stack([points[:, 0] + delta, heights_right, points[:, 1]], axis=1)
        points_up = np.stack([points[:, 0], heights_up, points[:, 1] + delta], axis=1)
        right_vector = points_right - points_3d
        up_vector = points_up - points_3d
        right_vector /= np.linalg.norm(right_vector, axis=1, keepdims=True)
        up_vector /= np.linalg.norm(up_vector, axis=1, keepdims=True)
        return cast(np.ndarray, np.cross(up_vector, right_vector))

    def create_edge_region(self, region: Region, points_per_meter: float, rand: np.random.Generator) -> EdgedRegion:
        grid_x, grid_y = self.region.get_linspace(self.cells_per_meter)
        interp_spline = RegularGridInterpolator((grid_x, grid_y), self.Z)

        # x/y between two points of the region are fully determined by the two endpoints
        region_points = region.points
        line_num = 0
        all_vertices = []
        for i, start_point in enumerate(region_points):
            end_point = region_points[(i + 1) % len(region_points)]
            meters = ((start_point[0] - end_point[0]) ** 2 + (start_point[1] - end_point[1]) ** 2) ** 0.5
            num_points = round(meters * points_per_meter) + 1
            points = np.linspace(start_point, end_point, num_points)

            if _LegacyEdgeNoise.IS_ENABLED:
                points = _LegacyEdgeNoise.add_legacy_edge_noise(points, line_num, points_per_meter, rand)
            heights = interp_spline(points)

            vertices = np.stack((points[:, 1], heights, points[:, 0]), axis=1)

            all_vertices.append(vertices[1:])

            line_num += 1

        all_vertices_np = np.concatenate(all_vertices, axis=0)

        if IS_DEBUG_VIS:
            plot_points(all_vertices, 0, 2)

        return EdgedRegion(region.x, region.z, all_vertices_np)

    def blur(self, blur_meters_std_dev: float) -> None:
        if blur_meters_std_dev > 0:
            blur = blur_meters_std_dev * self.cells_per_meter
            self.Z = gaussian_filter(self.Z, sigma=blur)

    def tilt(self, x_delta: float, y_delta: float) -> None:
        for i, j in np.ndindex(*self.Z.shape):
            x = self.X[i, j]
            y = self.Y[i, j]
            self.Z[i, j] += x * x_delta + y * y_delta

    def sink_edges(
        self,
        rand: np.random.Generator,
        is_circular: bool,
        fade_fraction: float,
        fade_noise_scale: float,
        fade_noise_min: float,
        extra_weight: float = 1.0,
    ) -> None:
        fade_weights = np.zeros_like(self.Z)
        assert fade_weights.shape[0] == fade_weights.shape[1], "Doesnt support rectangular arrays"

        mid_idx = np.clip(fade_weights.shape[0] // 2, 0, fade_weights.shape[0] - 1)
        points = self.get_2d_points()
        center_point = points[mid_idx, mid_idx]
        deltas = points - center_point
        if is_circular:
            dist = np.sqrt(deltas[:, :, 0] ** 2 + deltas[:, :, 1] ** 2)
        else:
            dist = np.max(np.abs(deltas), axis=2)
        edge_start_idx = np.clip(round((1.0 - fade_fraction) * fade_weights.shape[0]), 0, fade_weights.shape[0] - 1)
        edge_start_dist = dist[mid_idx, edge_start_idx]
        edge_end_dist = dist[mid_idx, -1]
        falloff_dist = edge_end_dist - edge_start_dist
        if falloff_dist < 0.001:
            return
        fade_weights = (dist - edge_start_dist) / falloff_dist
        fade_weights[dist < edge_start_dist] = 0.0
        # plot_value_grid(fade_weights)

        noise = perlin(fade_weights.shape, fade_noise_scale, rand)
        # plot_value_grid(noise)
        noise -= np.min(noise)
        noise /= np.max(noise) * (1 / fade_noise_min)
        noise += fade_noise_min

        fade_weights *= noise
        # plot_value_grid(fade_weights, title="FADE WEIGHTS")
        max_height = np.max(self.Z)
        self.Z -= max_height * fade_weights * extra_weight

    def lower_edges(self, final_height: float, fade_over_cells: int) -> None:
        Z2 = self.Z.copy()
        _min_helper(Z2[0, :], final_height)
        _min_helper(Z2[-1, :], final_height)
        _min_helper(Z2[:, 0], final_height)
        _min_helper(Z2[:, -1], final_height)
        for i in range(1, 1 + fade_over_cells):
            factor = i / fade_over_cells
            Z2[i, :] *= factor
            Z2[-1 + -i, :] *= factor
            Z2[:, i] *= factor
            Z2[:, -1 + -i] *= factor
        self.Z = Z2

    def get_elevation(self) -> MapFloatNP:
        return self.Z.copy()

    def get_squared_slope(self) -> MapFloatNP:
        """:returns: squared slope. Squared for efficiency (dont have to do as m any square roots"""
        meters_per_cell = 1.0 / self.cells_per_meter
        horizontal_slope = _calculate_horizontal_delta(self.Z) / (2 * meters_per_cell)
        vertical_slope = _calculate_vertical_delta(self.Z) / (2 * meters_per_cell)
        return cast(MapFloatNP, horizontal_slope * horizontal_slope + vertical_slope * vertical_slope)

    def _get_ocean_mask(self, water_mask: MapBoolNP) -> Optional[MapBoolNP]:
        for edge, ind in (
            (water_mask[0, :], (0, None)),
            (water_mask[-1, :], (-1, None)),
            (water_mask[:, 0], (None, 0)),
            (water_mask[:, -1], (None, -1)),
        ):
            ocean_indices = np.argwhere(edge)
            if len(ocean_indices) > 0:
                value = ocean_indices[0][0]
                ocean_index = tuple([(value if x is None else x) for x in ind])
                break
        # nothing to do if there is no ocean
        else:
            return None
        ocean_mask = water_mask.copy().astype(int)
        flood_fill(ocean_mask, ocean_index, 2, in_place=True)
        return cast(MapBoolNP, ocean_mask == 2)

    def add_debug_circle_lake(self, position: Point2DNP, radius: float, fade_radius: float) -> None:
        points = self.get_2d_points()
        deltas = points - position
        distances = np.linalg.norm(deltas, axis=2)
        weights = np.zeros_like(distances)
        fade_range = fade_radius - radius
        weights[distances < fade_radius] = (fade_radius - distances[distances < fade_radius]) / fade_range
        weights[distances < radius] = 1.0

        plot_value_grid(weights, "Debug circle lake (before)")

        height = -1.0
        land_mask = self.get_land_mask()
        self.Z[np.logical_and(land_mask, weights >= 1.0)] = height

        plot_value_grid(self.Z, "Debug circle lake (mid)")

        interpolation_mask = np.logical_and(land_mask, np.logical_and(weights < 1.0, weights > 0.0))
        self.Z[interpolation_mask] = (self.Z[interpolation_mask] - height) * (
            1.0 - weights[interpolation_mask]
        ) + height

        plot_value_grid(self.Z, "Debug circle lake (after)")

    def erode_shores(
        self,
        rand: np.random.Generator,
        biome_config: BiomeConfig,
        is_debug_graph_printing_enabled: bool = False,
    ) -> SpecialBiomes:
        """
        What is a beach, really?

        This function is of the opinion that a beach is:
        - On the ocean
        - Gently sloped
        - Sufficiently large

        Such regions are found, and then eroded (reduce land heights near the water for a flatter beach)

        :returns: two masks (for dark and light beach, ie, near and far from the water)
        """

        # figure out all shore points on the ocean
        water_mask = self.Z < WATER_LINE
        ocean_mask = cast(NDArray[np.bool_], self._get_ocean_mask(water_mask))
        fresh_water_mask = np.logical_and(water_mask, np.logical_not(ocean_mask))

        if not biome_config.is_beach_computed:
            beach_mask = np.zeros_like(fresh_water_mask)
            swamp_mask = np.zeros_like(fresh_water_mask)
        else:
            # get the beach mask, if any
            beach_mask = self._get_beach_mask(
                rand, biome_config, ocean_mask, water_mask, is_debug_graph_printing_enabled
            )

            # figure out the other biomes

            fresh_water_shore_mask = np.logical_and(
                morphology.dilation(fresh_water_mask, morphology.square(3)), np.logical_not(fresh_water_mask)
            )
            if np.any(fresh_water_shore_mask):
                tree = create_kd_tree(
                    np.argwhere(fresh_water_shore_mask),
                    "fresh water shore",
                    max_points=biome_config.max_kd_points,
                    rand=rand,
                )
                shore_distance = np.zeros_like(self.Z)
                is_land = self.Z >= WATER_LINE
                shore_distance[is_land] = tree.query(np.argwhere(is_land), k=1)[0]
                shore_distance = shore_distance * (1.0 / self.cells_per_meter)
                swamp_noise = (
                    perlin(shore_distance.shape, biome_config.swamp_noise_scale, rand)
                    * biome_config.swamp_noise_magnitude
                    * biome_config.swamp_distance_max
                )
                inland_area_below_swamp_line = np.logical_and(
                    self.Z < WATER_LINE + biome_config.swamp_elevation_max, np.logical_not(ocean_mask)
                )
                swamp_mask = np.logical_and(
                    inland_area_below_swamp_line, shore_distance < biome_config.swamp_distance_max + swamp_noise
                )
                if is_debug_graph_printing_enabled:
                    plot_value_grid(swamp_noise, "swamp noise")
                    plot_value_grid(
                        np.logical_and(inland_area_below_swamp_line, shore_distance < biome_config.swamp_distance_max),
                        "swamp without noise",
                    )
                    plot_value_grid(swamp_mask, "swamp with noise")
            else:
                swamp_mask = fresh_water_shore_mask.copy()

        return SpecialBiomes(beach_mask=beach_mask, swamp_mask=swamp_mask, fresh_water_mask=fresh_water_mask)

    def get_all_coast(self, extra_unclimbable_distance: float = 1.5) -> MapBoolNP:
        water_mask = self.Z < WATER_LINE
        shore_mask = np.logical_and(morphology.dilation(water_mask, morphology.square(3)), np.logical_not(water_mask))
        cell_distance = round(0.5 * extra_unclimbable_distance * self.cells_per_meter)
        return cast(MapBoolNP, morphology.dilation(shore_mask, morphology.disk(cell_distance)))

    def _get_beach_mask(
        self,
        rand: np.random.Generator,
        biome_config: BiomeConfig,
        ocean_mask: MapBoolNP,
        water_mask: MapBoolNP,
        is_debug_graph_printing_enabled: bool = True,
    ):
        if ocean_mask is None:
            empty_mask = np.zeros_like(water_mask)
            return empty_mask

        shore_mask = np.logical_and(morphology.dilation(ocean_mask, morphology.square(3)), np.logical_not(ocean_mask))

        if is_debug_graph_printing_enabled:
            plot_value_grid(ocean_mask, "ocean")
            plot_value_grid(shore_mask, "ocean shore")

        # nothing to do if there is no shore:
        if not np.any(shore_mask):
            empty_mask = np.zeros_like(water_mask)
            return empty_mask

        # compute slopes at shore points
        squared_slope = self.get_squared_slope()
        # remove shore points where the slope into the water is too high
        gently_sloped_shore_mask = np.zeros_like(shore_mask)
        gently_sloped_shore_mask[shore_mask] = squared_slope[shore_mask] < biome_config.beach_slope_cutoff

        if is_debug_graph_printing_enabled:
            plot_value_grid(gently_sloped_shore_mask, "gently sloped ocean shore")

        # expand the shore indices. This helps connect nearby regions
        expansion_cell_distance = round((biome_config.max_shore_distance / 2.0) * self.cells_per_meter)
        expanded_shore_mask = np.logical_and(
            shore_mask,
            morphology.dilation(gently_sloped_shore_mask, morphology.square(1 + (2 * expansion_cell_distance))),
        )

        if is_debug_graph_printing_enabled:
            plot_value_grid(expanded_shore_mask, "expanded shore mask")

        # contract the shore indices by more than we expanded. This prevents us from eating into nearby regions
        contraction_cell_distance = round(
            biome_config.beach_shrink_fraction * biome_config.max_shore_distance * self.cells_per_meter
        )
        rest_of_shore_mask = np.logical_and(shore_mask, np.logical_not(expanded_shore_mask))
        expanded_rest_of_shore_mask = np.logical_and(
            shore_mask,
            morphology.dilation(rest_of_shore_mask, morphology.square(1 + (2 * contraction_cell_distance))),
        )
        contracted_shore_mask = np.logical_and(expanded_shore_mask, np.logical_not(expanded_rest_of_shore_mask))

        if is_debug_graph_printing_enabled:
            plot_value_grid(contracted_shore_mask, "contracted shore mask")

        # create connected beach components
        lw, num = ndimage.label(contracted_shore_mask, structure=generate_binary_structure(2, 2))
        # remove beach components that are too small
        all_areas = ndimage.sum(contracted_shore_mask, lw, index=np.arange(np.max(lw) + 1))
        selected_area_ids = []
        min_cells_for_beach = biome_config.min_beach_length * self.cells_per_meter
        for area_id, area in enumerate(all_areas):
            if area < min_cells_for_beach:
                continue
            selected_area_ids.append(area_id)
        large_patch_shore_indices = np.argwhere(np.isin(lw, tuple(selected_area_ids), assume_unique=True))

        if is_debug_graph_printing_enabled:
            debug_large_patch_shore = np.zeros_like(self.Z)
            for index in list(large_patch_shore_indices):
                debug_large_patch_shore[tuple(index)] = 1.0
            plot_value_grid(debug_large_patch_shore, "large patch gentle slope ocean shore")

        # compute beach line distance matrix
        tree = create_kd_tree(
            large_patch_shore_indices, "large patch", max_points=biome_config.max_kd_points, rand=rand
        )
        shore_distance = np.zeros_like(self.Z)
        is_land = self.Z >= WATER_LINE
        shore_distance[is_land] = tree.query(np.argwhere(is_land), k=1)[0]
        shore_distance = shore_distance * (1.0 / self.cells_per_meter)

        if is_debug_graph_printing_enabled:
            plot_value_grid(shore_distance, title="shore distances")

        # the beach is anything below a (noisy) height AND within a (noisy) distance of the beach line
        beach_mask = np.logical_and(
            np.logical_and(self.Z < biome_config.max_shore_height, shore_distance < biome_config.max_shore_distance),
            is_land,
        )

        if is_debug_graph_printing_enabled:
            plot_value_grid(beach_mask, title="beach mask")

        # sometimes can end up with little islands of terrain isolated on the beach.
        # to deal with this, blend them in
        non_beach_land = np.logical_and(is_land, np.logical_not(beach_mask))
        # dilate once. this will be removed from the beach mask at the very end so that the region ids are set correctly
        beach_fade_cell_distance = round(0.5 * biome_config.beach_fade_distance * self.cells_per_meter)
        expanded_non_beach_land = morphology.dilation(non_beach_land, morphology.disk(beach_fade_cell_distance))
        # dilate again. the delta here defines the region over which we will blend heights
        full_non_beach_land = morphology.dilation(expanded_non_beach_land, morphology.disk(beach_fade_cell_distance))
        fade_region = np.logical_and(full_non_beach_land, np.logical_not(non_beach_land))
        fade_region = np.logical_and(fade_region, beach_mask)
        if np.any(fade_region):
            # figure out the distance to make this mask
            non_beach_edge = np.logical_and(
                morphology.dilation(non_beach_land, morphology.square(3)), np.logical_not(non_beach_land)
            )
            land_tree = create_kd_tree(
                np.argwhere(non_beach_edge), "non beach", max_points=biome_config.max_kd_points, rand=rand
            )
            land_distance = np.zeros_like(self.Z)
            land_distance[fade_region] = land_tree.query(np.argwhere(fade_region), k=1)[0]

            if np.max(land_distance) == 0:
                return beach_mask

            fade_weights = 1.0 - (land_distance / np.max(land_distance))
            fade_weights[np.logical_not(fade_region)] = 0.0

            if is_debug_graph_printing_enabled:
                plot_value_grid(fade_region, title="fade region")
                plot_value_grid(fade_weights, title="fade weights")

            # compute the "ideal" heights for each point on the beach (from the distance to the beach line)
            target_height = 0.1
            ideal_heights = target_height * shore_distance

            if is_debug_graph_printing_enabled:
                old_Z = self.Z.copy()

            # mix the ideal heights with the true heights
            mixing_weights = np.clip(shore_distance[beach_mask] / biome_config.max_shore_distance, 0, 1)
            mixing_weights = np.max(np.stack([fade_weights[beach_mask], mixing_weights], axis=1), axis=1)
            self.Z[beach_mask] = (
                mixing_weights * self.Z[beach_mask] + (1.0 - mixing_weights) * ideal_heights[beach_mask]
            )

            if is_debug_graph_printing_enabled:
                mixing_weights_debug = np.zeros_like(self.Z)
                mixing_weights_debug[beach_mask] = mixing_weights
                plot_value_grid(mixing_weights_debug, "mixing weights")
                plot_value_grid(self.Z, "new heights")
                plot_value_grid(self.Z - old_Z, "height delta")

            # finally, remove the stuff near the edges
            beach_mask = np.logical_and(beach_mask, np.logical_not(expanded_non_beach_land))

        if is_debug_graph_printing_enabled:
            plot_value_grid(beach_mask, "beach")

        return beach_mask

    def get_water_distance(
        self,
        rand: np.random.Generator,
        is_fresh_water_included_in_moisture: bool,
        max_points: Optional[int] = None,
        for_points: Optional[MapBoolNP] = None,
    ) -> Optional[MapFloatNP]:

        moisture = self.Z.copy()
        water_points = self.Z < WATER_LINE
        moisture[water_points] = 0.0
        land_points = self.Z >= WATER_LINE
        moisture[land_points] = 1.0

        if is_fresh_water_included_in_moisture:
            shore_points = _get_shore_points(rand, water_points)
        else:
            ocean_mask = self._get_ocean_mask(water_points)
            if ocean_mask is None:
                return None
            shore_points = _get_shore_points(rand, ocean_mask)

        num_points = len(shore_points)
        if num_points == 0:
            return None
        tree = create_kd_tree(shore_points, "shore", max_points=max_points, rand=rand)

        if for_points is not None:
            result = np.zeros_like(moisture)
            result[np.nonzero(for_points)] = tree.query(np.argwhere(for_points), k=1)[0]
            return result * (1.0 / self.cells_per_meter)

        moisture[np.nonzero(moisture)] = tree.query(np.argwhere(moisture), k=1)[0]
        return moisture * (1.0 / self.cells_per_meter)

    def distances_from_points(
        self,
        point_indices: np.ndarray,
        mask: MapBoolNP,
        name: str,
        max_points: Optional[int] = None,
        rand: Optional[np.random.Generator] = None,
    ):
        distances = np.zeros_like(self.Z)
        tree = create_kd_tree(point_indices, name, max_points, rand)
        distances[np.nonzero(mask)] = tree.query(np.argwhere(mask), k=1)[0]
        return distances * (1.0 / self.cells_per_meter)

    def _full_distances_from_points(
        self,
        point_indices: np.ndarray,
        mask: MapBoolNP,
        name: str,
        max_points: Optional[int] = None,
        rand: Optional[np.random.Generator] = None,
    ):
        distances = np.zeros_like(self.Z)
        tree = create_kd_tree(point_indices, name, max_points, rand)
        distances[np.nonzero(mask)], indices = tree.query(np.argwhere(mask), k=1)
        return distances * (1.0 / self.cells_per_meter), indices

    def point_to_index(self, point: Point2DNP) -> SupportsIndexTuple:
        assert_isinstance(point, Point2DNP)
        max_x_index = self.Z.shape[1] - 1
        one_cell_x = self.region.x.size / (max_x_index)
        x = round((point[0] - self.region.x.min_ge) / one_cell_x)
        max_y_index = self.Z.shape[0] - 1
        one_cell_y = self.region.z.size / (max_y_index)
        y = round((point[1] - self.region.z.min_ge) / one_cell_y)
        return cast(SupportsIndexTuple, (clamp(y, 0, max_y_index), clamp(x, 0, max_x_index)))

    def index_to_point_2d(self, indices: Tuple[int, int]) -> Point2DNP:
        return np.array([self.X[indices], self.Y[indices]])

    def raise_island(self, island: MapBoolNP, delta: float) -> None:
        self.Z[island] += delta

    def lower_island(self, island: MapBoolNP, set_to_height: float):
        min_height = self.Z[island].min()
        amount = min_height - set_to_height
        self.Z[island] -= amount
        return amount

    def get_outline(self, mask: MapBoolNP, cell_thickness: int):
        outline = np.logical_and(morphology.dilation(mask, morphology.disk(1)), np.logical_not(mask))
        if cell_thickness > 1:
            outline = morphology.dilation(outline, morphology.disk(cell_thickness))
        return outline

    def shrink_region(self, mask: MapBoolNP, num_cells: int):
        return np.logical_and(
            mask, np.logical_not(morphology.dilation(np.logical_not(mask), morphology.disk(num_cells)))
        )

    def interpolate_heights(self, region: MapBoolNP, border: MapBoolNP) -> None:
        expanded_region = morphology.dilation(region, morphology.disk(1))
        border = np.logical_and(expanded_region, border)

        # plot_value_grid(region, "region")
        # plot_value_grid(border, "border")

        point_ind = np.argwhere(border)
        border_dist, border_ind = self._full_distances_from_points(point_ind, region, "border a")
        border_dist = border_dist[region]
        near_point_inds = point_ind[border_ind]
        heights = self.Z[near_point_inds[:, 0], near_point_inds[:, 1]]

        # TODO: fix this--should be based on how much space we have to the safety radius, really
        max_dist = 10.0
        mix_factor = np.clip((2 * border_dist / max_dist) - 1.0, 0, 1)

        # old_z = self.Z.copy()

        self.Z[region] = mix_factor * self.Z[region] + (1.0 - mix_factor) * heights

        # plot_value_grid(self.Z - old_z)

    def log_simplicity_warning(self, message: str) -> None:
        self._simplicity_warnings.append(message)

    def clear_simplicity_warnings(self) -> None:
        self._simplicity_warnings.clear()

    def get_simplicity_warnings(self) -> Tuple[str, ...]:
        return tuple(self._simplicity_warnings)

    def create_radial_mask(self, point: Point2DNP, radius: float) -> MapFloatNP:
        assert_isinstance(point, Point2DNP)
        dist_sq = self.get_dist_sq_to(point)

        result = np.zeros_like(self.Z)
        mask = dist_sq < radius * radius
        result[mask] = np.clip(((radius - np.sqrt(dist_sq[mask])) / radius) * 2, 0, 1)

        # plot_value_grid(result)

        return result

    def add_hill(self, point: Point2DNP, scale: float, radius: float, mask: MapBoolNP) -> None:
        assert_isinstance(point, Point2DNP)
        dist_sq = self.get_dist_sq_to(point)

        result = np.zeros_like(self.Z)
        nearby_mask = np.logical_and(dist_sq < radius * radius, mask)
        result[nearby_mask] = np.clip(((radius - np.sqrt(dist_sq[nearby_mask])) / radius), 0, 1)
        # plot_value_grid(result)
        temp = self.Z[nearby_mask] * scale
        self.Z[nearby_mask] = (1.0 - result[nearby_mask]) * self.Z[nearby_mask] + result[nearby_mask] * temp


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FrozenHeightMap(HeightMap, Serializable):
    _simplicity_warnings: Tuple[str, ...]  # type: ignore


def create_kd_tree(
    points: np.ndarray, name: str, max_points: Optional[int] = None, rand: Optional[np.random.Generator] = None
):
    reduction = 1.0
    if max_points is not None:
        reduction = max_points / len(points)
        if reduction < 1.0:
            assert rand is not None
            points = np.logical_and(points, rand.uniform(size=points.shape) < reduction)
    # logger.debug(f"Creating {name} KDTree with {round(len(points) * reduction)} points")
    return KDTree(points)


def _get_shore_points(rand: np.random.Generator, water_points: np.ndarray):
    border, border_points = get_border_points(water_points, rand)
    return border_points


def _min_helper(value: np.ndarray, min_value: float) -> None:
    value[value > min_value] = min_value


NumberType = TypeVar("NumberType", bound=Union[int, float, np.floating])


def clamp(n: NumberType, smallest: NumberType, largest: NumberType) -> NumberType:
    return max(smallest, min(n, largest))  # type: ignore


def _calculate_horizontal_delta(Z: np.ndarray):
    first_col = np.expand_dims(Z[:, 0] - (Z[:, 1] - Z[:, 0]), 1)
    last_col = np.expand_dims(Z[:, -1] - (Z[:, -2] - Z[:, -1]), 1)
    combined = np.concatenate([first_col, Z, last_col], axis=1)
    return combined[:, 2:] - combined[:, :-2]


def _calculate_vertical_delta(Z: np.ndarray):
    first_row = np.expand_dims(Z[0, :] - (Z[1, :] - Z[0, :]), 0)
    last_row = np.expand_dims(Z[-1, :] - (Z[-2, :] - Z[-1, :]), 0)
    combined = np.concatenate([first_row, Z, last_row], axis=0)
    return combined[2:, :] - combined[:-2, :]


def lineseg_dists(p: np.ndarray, a: np.ndarray, b: np.ndarray):
    """Cartesian distance from point to line segment

    Edited to support arguments as series, from:
    https://stackoverflow.com/a/54442561/11208892

    Args:
        - p: np.array of single point, shape (2,) or 2D array, shape (x, 2)
        - a: np.array of shape (x, 2)
        - b: np.array of shape (x, 2)
    """
    # normalized tangent vectors
    d_ba = b - a
    # if these two points are the same, we just care about distance from a point
    if abs(d_ba.sum()) < 0.001:
        return np.linalg.norm(p - a, axis=1)
    d = np.divide(d_ba, (np.hypot(d_ba[:, 0], d_ba[:, 1]).reshape(-1, 1)))
    # signed parallel distance components
    # rowwise dot products of 2D vectors
    s = np.multiply(a - p, d).sum(axis=1)
    t = np.multiply(p - b, d).sum(axis=1)
    # clamped parallel distance
    h = np.maximum.reduce([s, t, np.zeros(len(s))])
    # perpendicular distance component
    # rowwise cross products of 2D vectors
    d_pa = p - a
    c = d_pa[:, 0] * d[:, 1] - d_pa[:, 1] * d[:, 0]
    return np.hypot(h, c)


def get_border_points(data: np.ndarray, rand: np.random.Generator, reduction: float = 0.3):
    data = data.astype(int)
    kernel = np.array(
        [
            [0, 1, 0],
            [16, 0, 16**2],
            [0, 16**3, 0],
        ]
    )
    sums = convolve(data, kernel)
    expected_sums = data + data * 16 + data * 16**2 + data * 16**3
    border = sums != expected_sums
    # plot_value_grid(border, "BORDER")
    # make it go 3x faster and add a little noise...  we are bad people
    # might even get limited later as well... (by the KD tree construction)
    border = np.logical_and(border, rand.uniform(size=border.shape) < reduction)
    border_points = np.argwhere(border)
    return border, border_points


def selected_distance_weighted_points(
    rand: np.random.Generator,
    possible_points: Point2DListNP,
    center_point: Point2DNP,
    goal_distance_weight: stats.norm,
    point_count: int,
):
    distances = np.linalg.norm(possible_points - center_point, axis=1)
    location_weights = goal_distance_weight.pdf(distances)
    location_probabilities = location_weights / location_weights.sum()
    return rand.choice(possible_points, point_count, p=location_probabilities)


def get_flora_config_by_file(
    flora_config: Dict[str, FloraConfig], resource_file: str
) -> Tuple[Optional[FloraConfig], str]:
    resource_name = resource_file.split("/")[-1].split(".")[0]
    if resource_name in flora_config:
        config: Optional[FloraConfig] = flora_config[resource_name]
    else:
        config = flora_config.get(resource_name.replace("_low", ""), None)
    return config, resource_name


def build_outdoor_world_map(config: WorldConfig, is_debug_graph_printing_enabled: bool = False) -> HeightMap:
    """
    Fractal generation of HeightMap's

    Does the same operations (add noise, raise interior random parts of the terrain, upsample) repeatedly.

    At the end, does some smoothing and rescales to the proper height.
    """

    # create a flat heightmap of the correct size
    region = Region(
        FloatRange(-config.size_in_meters / 2.0, config.size_in_meters / 2.0),
        FloatRange(-config.size_in_meters / 2.0, config.size_in_meters / 2.0),
    )
    cells_per_meter = config.initial_point_count / config.size_in_meters
    map = HeightMap.create(region, cells_per_meter)

    rand = np.random.default_rng(config.seed)

    # set initial scale parameters
    noise = config.size_in_meters * config.initial_noise_scale
    mountain_radius = config.mountain_radius

    # fractal iteration--keep doing the same thing, but reducing the scale of the operations
    for i in range(config.fractal_iteration_count):

        # add some noise to the existing points
        map.add_noise(noise, rand)
        if is_debug_graph_printing_enabled:
            map.plot()

        # if we're adding mountains, add them
        if config.mountain_noise_count:
            for j in range(config.mountain_noise_count):
                if config.is_mountain_placement_normal_distribution:
                    x_remaining_space = (map.region.x.size / 2.0) - config.mountain_offset
                    z_remaining_space = (map.region.x.size / 2.0) - config.mountain_offset
                    mountain_center = (
                        rand.normal(map.region.x.midpoint, x_remaining_space / 2.0),
                        rand.normal(map.region.z.midpoint, z_remaining_space / 2.0),
                    )

                else:
                    offset = map.region.x.size * config.mountain_offset
                    mountain_center = (
                        rand.uniform(map.region.x.min_ge + offset, map.region.x.max_lt - offset),
                        rand.uniform(map.region.z.min_ge + offset, map.region.z.max_lt - offset),
                    )
                map.add_center_biased_noise(
                    mountain_center, noise * config.mountain_noise_scale, mountain_radius, rand
                )
            if is_debug_graph_printing_enabled:
                map.plot()

        # upsample the grid (make it finer resolution so the next pass works at a smaller scale)
        map = map.upsample()
        if is_debug_graph_printing_enabled:
            map.plot()

        # decrease the scale parameters
        mountain_radius *= config.mountain_radius_decay
        noise *= config.noise_scale_decay

    map.blur(config.blur_meters_std_dev)

    if is_debug_graph_printing_enabled:
        map.plot()

    # first do normal fade with configuration as a circle
    map.sink_edges(
        rand, config.is_fade_circular, config.fade_fraction, config.fade_noise_scale, config.fade_noise_min, 5.0
    )
    # last ditch fade out the rest
    map.sink_edges(rand, False, config.fade_fraction / 10.0, config.fade_noise_scale, 0.9)

    # scale our heights to be what was requested
    max_altitude = map.Z.max(initial=0.1)
    map.Z *= config.final_max_altitude_meters / max_altitude

    # # this is a bit of a hack. We force the edges of the world down so they're guaranteed to be under water
    # # this mostly doesn't matter because the above, more elegant ways of putting the edges of the world underwater usually do the trick
    # # but this makes absolutely certain that every edge is underwater
    # final_edge_height = -1.0
    # cells_over_which_to_fade = round((map.X.shape[0]) * config.final_world_edge_fade_fraction) + 1
    # map.lower_edges(final_edge_height, cells_over_which_to_fade)

    # force the edges of the world to be underwater, no mater what:
    min_water_border_cells = 4
    map.Z[:, 0:min_water_border_cells] = np.clip(map.Z[:, 0:min_water_border_cells], None, -1.0)
    map.Z[:, -min_water_border_cells:] = np.clip(map.Z[:, -min_water_border_cells:], None, -1.0)
    map.Z[0:min_water_border_cells, :] = np.clip(map.Z[0:min_water_border_cells, :], None, -1.0)
    map.Z[-min_water_border_cells:, :] = np.clip(map.Z[-min_water_border_cells:, :], None, -1.0)

    if is_debug_graph_printing_enabled:
        map.plot()

    return map


# TODO(mjr) added back for regression tests. Should probably disentangle them from world generation
class _LegacyEdgeNoise:
    IS_ENABLED: ClassVar = False

    @staticmethod
    def add_legacy_edge_noise(
        points: np.ndarray,
        line_num: int,
        points_per_meter: float,
        rand: np.random.Generator,
        noise_scale: float = 0.5,
    ) -> np.ndarray:
        meters_per_point = 1.0 / points_per_meter
        width = noise_scale * (meters_per_point * 0.5) * 0.999
        noise = rand.uniform(-width, width, points.shape)
        if line_num in (1, 3):
            noise[:, 0] = 0.0
        else:
            noise[:, 1] = 0.0
        noise[0, 0] = 0.0
        noise[0, 1] = 0.0
        noise[-1, 0] = 0.0
        noise[-1, 1] = 0.0

        # TODO: put back, if I can get the random generation working properly...
        # noised_points = points + noise
        points[:, [1, 0]] = points[:, [0, 1]]
        return points



from typing import List

import attr
import numpy as np
from godot_parser import Node as GDNode
from godot_parser import NodePath

from avalon.common.utils import only
from avalon.datagen.world_creation.entities.doors.door import Door
from avalon.datagen.world_creation.entities.doors.locks.rotating_bar import RotatingBar
from avalon.datagen.world_creation.entities.doors.locks.sliding_bar import SlidingBar
from avalon.datagen.world_creation.entities.doors.types import LatchingMechanics
from avalon.datagen.world_creation.indoor.blocks import new_make_block_nodes
from avalon.datagen.world_creation.indoor.utils import get_scaled_mesh_node
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.utils import make_transform


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class SlidingDoor(Door):
    slide_right: bool = True
    handle_size_proportion: float = 0.2
    latching_mechanics: LatchingMechanics = LatchingMechanics.NO_LATCH

    def get_node(self, scene: GodotScene) -> GDNode:
        DOOR_BODY = "body"

        body_nodes = self._get_body_nodes(scene, DOOR_BODY)
        rail_nodes = self._get_rail_nodes(scene, DOOR_BODY)

        door_script = scene.add_ext_resource("res://entities/doors/sliding_door.gd", "Script")
        group_spatial_node = GDNode(
            f"sliding_door_{self.entity_id}",
            "Spatial",
            properties={
                "transform": make_transform(self.rotation, self.position),
                "script": door_script.reference,
                "entity_id": self.entity_id,
                "slide_axis": self.face_axis.value,
                "is_latched": self.latching_mechanics != LatchingMechanics.NO_LATCH,
            },
        )
        for body_node in body_nodes:
            group_spatial_node.add_child(body_node)
        for rail_node in rail_nodes:
            group_spatial_node.add_child(rail_node)

        body_node = only(body_nodes)
        for i, lock in enumerate(self.locks):
            # if we have two locks of the same kind, their instance names will otherwise clash
            lock_with_id = attr.evolve(lock, entity_id=i)
            group_spatial_node.add_child(lock_with_id.get_node(scene))
            for body_addition in lock_with_id.get_additional_door_body_nodes(scene, self.size):
                body_node.add_child(body_addition)
        return group_spatial_node

    def _get_body_nodes(self, scene: GodotScene, door_body_node_name: str) -> List[GDNode]:
        body_children_nodes = new_make_block_nodes(
            scene,
            np.array([0, 0, 0]),
            self.size,
            make_parent=False,
            make_mesh=False,
            collision_shape_name=f"{door_body_node_name}_collision_shape",
        )
        door_body_mesh = get_scaled_mesh_node(
            scene, "res://entities/doors/door_body.tscn", self.size, mesh_name="body_mesh"
        )
        body_children_nodes.append(door_body_mesh)
        handle_nodes = self._get_handle_nodes(scene)

        door_width, door_height, door_thickness = self.size
        rail_hinge_distance = door_width * 0.5
        rail_hinge_width, rail_hinge_height, rail_hinge_thickness = 0.12, 0.32, 0.05
        rail_hinge_size = np.array([rail_hinge_width, rail_hinge_height, rail_hinge_thickness])
        rail_hinge_offset = (door_thickness - rail_hinge_thickness) / 2

        left_hinge_mesh_node = get_scaled_mesh_node(
            scene,
            "res://entities/doors/rail_hinge.tscn",
            rail_hinge_size,
            mesh_position=np.array([-rail_hinge_distance / 2, door_height / 2, rail_hinge_offset]),
            mesh_name="left_hinge_mesh",
        )
        body_children_nodes.append(left_hinge_mesh_node)

        right_hinge_mesh_node = get_scaled_mesh_node(
            scene,
            "res://entities/doors/rail_hinge.tscn",
            rail_hinge_size,
            mesh_position=np.array([rail_hinge_distance / 2, door_height / 2, rail_hinge_offset]),
            mesh_name="right_hinge_mesh",
        )
        body_children_nodes.append(right_hinge_mesh_node)

        body_script = scene.add_ext_resource("res://entities/doors/door_body.gd", "Script")
        body_node = GDNode(
            door_body_node_name,
            type="RigidBody",
            properties={
                "transform": make_transform(),
                "mass": 100.0,
                "script": body_script.reference,
                "entity_id": 0,
                "is_auto_latching": self.latching_mechanics == LatchingMechanics.AUTO_LATCH,
            },
        )
        for node in body_children_nodes:
            body_node.add_child(node)
        for node in handle_nodes:
            body_node.add_child(node)
        return [body_node]

    def _get_handle_nodes(self, scene: GodotScene) -> List[GDNode]:
        DOOR_HANDLE = "handle"

        door_width, door_height, door_thickness = self.size
        handle_width = door_width * (self.handle_size_proportion / 2)
        handle_height = door_height * self.handle_size_proportion
        handle_thickness = door_thickness * 0.75

        # We cap the handle height if there are locks that would get obstructed by it
        leeway = 0.05
        bar_locks = [lock for lock in self.locks if isinstance(lock, (RotatingBar, SlidingBar))]
        if len(bar_locks) > 0:
            max_y = np.inf
            min_y = -np.inf
            for lock in bar_locks:
                if not isinstance(lock, (RotatingBar, SlidingBar)):
                    continue
                if lock.position[1] > 0 and (lock_bottom := lock.position[1] - lock.size[1] / 2) < max_y:
                    max_y = lock_bottom
                elif lock.position[1] < 0 and (lock_top := lock.position[1] + lock.size[1] / 2) > min_y:
                    min_y = lock_top
            max_handle_height = min(abs(max_y), abs(min_y)) * 2 - leeway
            if handle_height > max_handle_height:
                handle_height = max_handle_height
        handle_size = np.array([handle_width, handle_height, handle_thickness])
        handle_margin_skew = 25  # left margin / right margin
        handle_side_margin = (door_width - handle_width) / (1 + handle_margin_skew)
        if not self.slide_right:
            handle_side_margin *= handle_margin_skew

        handle_nodes = []
        handle_position = np.array(
            [
                -door_width / 2 + handle_side_margin + handle_width / 2,
                0,
                (handle_thickness / 2 + door_thickness / 2),
            ]
        )
        handle_name = f"{DOOR_HANDLE}_0"
        handle_nodes.extend(
            new_make_block_nodes(
                scene,
                handle_position,
                handle_size,
                make_parent=False,
                make_mesh=False,
                collision_shape_name=f"{handle_name}_collision_shape",
            )
        )
        handle_mesh_node = get_scaled_mesh_node(
            scene,
            "res://entities/doors/door_handle_vertical.tscn",
            handle_size,
            handle_position,
            mesh_name=f"{handle_name}_mesh",
        )
        handle_nodes.append(handle_mesh_node)
        return handle_nodes

    def _get_rail_nodes(self, scene: GodotScene, door_body_node_name: str) -> List[GDNode]:
        DOOR_RAIL = "rail"

        door_width, door_height, door_thickness = self.size
        rail_height = door_height * 0.05
        rail_size = np.array([door_width * 2, rail_height, door_thickness * 1.1])
        direction_multiplier = 1 if self.slide_right else -1
        rail_position = np.array([direction_multiplier * door_width / 2, door_height / 2 + rail_height / 2, 0])

        rail_mesh_node = get_scaled_mesh_node(scene, "res://entities/doors/rail.tscn", rail_size)
        rail_node = only(
            new_make_block_nodes(
                scene,
                rail_position,
                rail_size,
                parent_name=DOOR_RAIL,
                make_mesh=False,
                collision_shape_disabled=True,
            )
        )
        rail_node.add_child(rail_mesh_node)

        slider_joint_position = direction_multiplier * door_width, door_height / 2, 0
        slider_joint_node = GDNode(
            "slider_joint",
            "SliderJoint",
            properties={
                "transform": make_transform(position=slider_joint_position),
                "nodes/node_a": NodePath(f"../{DOOR_RAIL}"),
                "nodes/node_b": NodePath(f"../{door_body_node_name}"),
                "linear_limit/lower_distance": 0 if self.slide_right else -door_width,
                "linear_limit/upper_distance": door_width if self.slide_right else 0,
                "linear_limit/softness": 0.5,
            },
        )
        return [rail_node, slider_joint_node]



import attr
import torch
from loguru import logger

from avalon.agent.common.params import ClippedNormalMode
from avalon.agent.common.parse_args import parse_args
from avalon.agent.common.trainer import OnPolicyTrainer
from avalon.agent.common.util import setup_new_process
from avalon.agent.godot.godot_eval import test
from avalon.agent.godot.godot_gym import GodotEnvironmentParams
from avalon.agent.godot.godot_gym import TrainingProtocolChoice
from avalon.agent.ppo.params import PPOParams
from avalon.common.log_utils import configure_remote_logger


@attr.s(auto_attribs=True, frozen=True)
class AvalonPPOParams(PPOParams):
    total_env_steps: int = 50_000_000
    num_steps: int = 200
    num_workers: int = 16
    batch_size: int = 200 * 16  # must equal num_steps * num_workers
    ppo_epochs: int = 2
    discount: float = 0.99
    lam: float = 0.83
    value_loss_coef: float = 1
    entropy_coef: float = 1.5e-4
    clip_range: float = 0.03
    lr: float = 2.5e-4
    clip_grad_norm: float = 0.5
    env_params: GodotEnvironmentParams = GodotEnvironmentParams(
        # if 100% success, difficulty in N steps will be N / (avg_ep_len * num_workers * num_tasks) * diff_update
        # eg 10M steps / (200 steps * 16 workers * 16 tasks) * 3e-4 = ~.06 max achievable difficulty
        task_difficulty_update=3e-4,
        meta_difficulty_update=3e-5,
        is_meta_curriculum_used=False,
        energy_cost_coefficient=1e-8,
        training_protocol=TrainingProtocolChoice.MULTI_TASK_BASIC,
        test_episodes_per_task=101,
        fixed_world_max_difficulty=0.5,
    )
    log_freq_hist: int = 500
    log_freq_scalar: int = 50
    log_freq_media: int = 500
    checkpoint_every: int = 2500

    eval_exploration_mode: str = "eval"  # TODO: try explore here also

    # TODO: should try converting this to the TruncatedNormal one
    policy_normal_init_std: float = 1
    policy_normal_min_std: float = 0.01
    clipped_normal_mode: ClippedNormalMode = ClippedNormalMode.NO_CLIPPING
    center_and_clamp_discrete_logits = True


def run(params: AvalonPPOParams) -> None:
    assert params.num_batches == 1
    assert params.batch_size == params.num_steps * params.num_workers

    trainer = OnPolicyTrainer(params)
    try:
        if params.is_training:
            trainer.train()
            trainer.shutdown()
            if trainer.train_storage:
                trainer.train_storage.reset()
            torch.cuda.empty_cache()  # just for seeing what's going on
        if params.is_testing:
            test(trainer.params, trainer.algorithm, log=True)

    finally:
        trainer.shutdown()


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    try:
        default_params = AvalonPPOParams()
        default_params = parse_args(default_params)
        run(default_params)
    except Exception as e:
        logger.exception(e)
        raise



from pathlib import Path

import attr
import numpy as np
from scipy import stats

from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import HALF_AGENT_HEIGHT_VECTOR
from avalon.datagen.world_creation.constants import TIGHT_DIST_STD_DEV
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.entities.constants import CANONICAL_FOOD_HEIGHT_ON_TREE
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.export import export_world


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class ScrambleTaskConfig(TaskConfig):
    # how steep the world will be, in general, between the spawn and goal
    # needs to be relatively close to this value, otherwise the task is simply walk or climb...
    desired_slope: float = 0.8
    # how far away the goal should be for difficulty 0.0 and 1.0 respectively
    goal_dist_easy: float = 2.0
    goal_dist_hard: float = 20.0
    # how "diverse" the world will be, ie, how crazy the terrain will be
    world_diversity_easy: float = 0.5
    world_diversity_hard: float = 1.0
    # worlds will be at least this large
    world_size_easy: float = 40.0
    world_size_hard: float = 60.0


def generate_scramble_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: ScrambleTaskConfig = ScrambleTaskConfig(),
) -> None:
    world_diversity = scale_with_difficulty(
        difficulty, task_config.world_diversity_easy, task_config.world_diversity_hard
    )
    world_min_size = scale_with_difficulty(difficulty, task_config.world_size_easy, task_config.world_size_hard)
    desired_goal_dist = scale_with_difficulty(difficulty, task_config.goal_dist_easy, task_config.goal_dist_hard)
    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, TIGHT_DIST_STD_DEV),
        rand,
        world_diversity,
        export_config,
        None,
        min_size_in_meters=world_min_size,
    )
    world, locations = world.end_height_obstacles(locations, is_accessible_from_water=False)

    # boost up the terrain around the food
    horizontal_distance = locations.get_2d_spawn_goal_distance()
    spawn_height = locations.spawn[1] - HALF_AGENT_HEIGHT_VECTOR[1]
    desired_height = horizontal_distance * task_config.desired_slope + spawn_height
    height_at_food = locations.goal[1] - CANONICAL_FOOD_HEIGHT_ON_TREE
    actual_height_delta = desired_height - height_at_food
    height_scale = desired_height / height_at_food
    map_new = world.map.copy()
    map_new.add_hill(to_2d_point(locations.goal), height_scale, horizontal_distance, locations.island)

    # flatten the top just a little for the purposes of tree placement
    flatten_size = min([horizontal_distance * 0.2, 2.0])
    map_new.radial_flatten(to_2d_point(locations.goal), flatten_size * 2.0)

    world = attr.evolve(world, map=map_new)

    # adjust the final food height by the delta
    locations = attr.evolve(locations, goal=locations.goal + UP_VECTOR * actual_height_delta)

    world = add_food_tree_for_simple_task(world, locations)
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    export_world(output_path, rand, world)



from avalon.datagen.world_creation.constants import AvalonTask


class DatagenConfigError(Exception):
    """
    Raised when datagen configuration values are impossible or incoherent.
    Ideally we keep the usage of this class to a minimum and use the type signatures to verify, but some
    combinations of invalid parameter values are hard to specify via the types.
    """


class GodotError(Exception):
    """
    Raised when the godot binary fails for any reason.
    """


class DesynchronizedTransformError(Exception):
    """
    Raised when your transforms are not perfectly synchronous.
    Is here to help keep you from going insane if you were to accidentally apply different transforms to
    data that really needs to stay synced (eg, depth and rgb)
    """


class ImpossibleWorldError(Exception):
    pass


class WorldTooSmall(ImpossibleWorldError):
    def __init__(self, task: AvalonTask, min_dist: float, available_dist: float, *args: object) -> None:
        super().__init__(f"Small {task} world: needed {min_dist} but only have {available_dist}", *args)



import sys
import typing
from enum import EnumMeta
from typing import Any
from typing import Type
from typing import TypeVar
from typing import Union

import attr

from avalon.agent.common.params import Params

ParamType = TypeVar("ParamType", bound=Params)


def strip_optional(field: Type) -> Type:
    # Optional is an alias for Union[x, None].
    # Return x if this is indeed an Optional, otherwise return field unchanged.
    origin = typing.get_origin(field)
    args = typing.get_args(field)
    if origin is Union and type(None) in args and len(args) == 2:
        assert args[1] == type(None)
        return typing.cast(Type, args[0])
    return field


def cast_argument_to_paramater_type(key: str, value: str, params: Any) -> Any:
    """Looks up the key in `params` and casts `value` to the type-hinted type of that parameter."""
    fields = attr.fields_dict(type(params))
    assert key in fields, f"could not find {key} in param object {params}."
    dtype = fields[key].type
    if dtype is None:
        raise ValueError("got none dtype - not sure what this means")
    dtype = strip_optional(dtype)
    if type(dtype) == type:
        if dtype == bool:
            if value.lower() == "false":
                return False
                # out[arg] = False
            elif value.lower() == "true":
                return True
                # out[arg] = True
            else:
                raise ValueError(f"could not parse {value} to bool")
        else:
            return dtype(value)
    elif type(dtype) == EnumMeta:
        return dtype[value]
    elif dtype == list[str]:
        # use comma-separated list with no spaces
        return value.split(",")
    elif dtype == tuple[str, ...]:
        # use comma-separated list with no spaces
        return tuple(value.split(","))
    elif dtype == tuple[int, ...]:
        # use comma-separated list with no spaces
        return tuple(value.split(","))
    else:
        assert type(dtype) == str, (key, dtype, type(dtype))
        # We got a forward type declaration, either from an explicit quote-wrapped type, or
        # `from __future__ import annotations`, or by default in future python.
        # In this case dtype will be a string, eg "int" instead of the actual type.
        # In theory attrs.resolve_types or typing.get_type_hints could be used to resolve these, but this breaks
        # if the class has a type defined with a TYPE_CHECKING guard (https://github.com/python/cpython/issues/87629).
        # Using `eval` is one way around this (and suggested in PEP563) but feels unsafe.
        # Probably better to just handle each case manually, eg `if dtype == "str": ...`.
        raise NotImplementedError


def recursive_fill_nested_argument(key: str, value: str, params: Any) -> Any:
    key_parts = key.split(".")
    if len(key_parts) == 1:
        key = key_parts[0]
        value = cast_argument_to_paramater_type(key, value, params)
        return attr.evolve(params, **{key: value})
    else:
        current_key = key_parts[0]
        child_params = params.__getattribute__(current_key)
        updated_params = recursive_fill_nested_argument(".".join(key_parts[1:]), value, child_params)
        return attr.evolve(params, **{current_key: updated_params})


def parse_args(params: ParamType) -> ParamType:
    """Parse args of the form `test.py --arg1 1 --arg2.child_arg value`.

    Args with part1.part2 are interpreted as args into nested params."""
    args = [x.strip() for x in " ".join(sys.argv[1:]).split("--") if x != ""]
    for arg in args:
        key, value = arg.split(" ")
        params = recursive_fill_nested_argument(key, value, params)
    return params



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The environment class for MonoBeast."""
import gym.spaces.utils
import torch


def _format_frame(frame):
    frame = torch.from_numpy(frame)
    return frame.view((1, 1) + frame.shape)  # (...) -> (T,B,...).


class Environment:
    def __init__(self, gym_env: gym.Env, device: str = "cuda:0") -> None:
        self.gym_env = gym_env
        self.episode_return = None
        self.episode_step = None
        self.device = device

        last_action_sample = gym_env.action_space.sample()
        self.default_action = torch.from_numpy(
            gym.spaces.utils.flatten(gym_env.action_space, last_action_sample)
        ).unsqueeze(0)

    def initial(self):
        initial_reward = torch.zeros(1, 1).to(self.device)
        # This supports only single-tensor actions ATM.

        # initial_last_action = torch.zeros(1, self.flat_action_dim, dtype=torch.float32)
        # initial_last_action[0, 0] = 1  # TODO unyuck
        initial_last_action = self.default_action.clone().to(self.device)

        self.episode_return = torch.zeros(1, 1).to(self.device)
        self.episode_step = torch.zeros(1, 1, dtype=torch.int32).to(self.device)
        initial_done = torch.ones(1, 1, dtype=torch.uint8).to(self.device)
        initial_frame = _format_frame(self.gym_env.reset()).to(self.device)
        return dict(
            frame=initial_frame,
            reward=initial_reward,
            done=initial_done,
            episode_return=self.episode_return,
            episode_step=self.episode_step,
            last_action=initial_last_action,
        )

    def step(self, action):
        frame, reward, done, info = self.gym_env.step(action)
        self.episode_step += 1
        self.episode_return += reward
        episode_step = self.episode_step
        episode_return = self.episode_return
        if done:
            frame = self.gym_env.reset()
            self.episode_return = torch.zeros(1, 1).to(self.device)
            self.episode_step = torch.zeros(1, 1, dtype=torch.int32).to(self.device)

        frame = _format_frame(frame).to(self.device)
        reward = torch.tensor(reward).view(1, 1).to(self.device)
        done = torch.tensor(done).view(1, 1).to(self.device)

        return dict(
            frame=frame,
            reward=reward,
            done=done,
            episode_return=episode_return,
            episode_step=episode_step,
            last_action=action,
            info=info,
        )

    def close(self) -> None:
        self.gym_env.close()



import os
from pathlib import Path
from typing import Tuple

import gym
from torch import Tensor

from avalon.agent.common.params import Params
from avalon.agent.common.trainer import OnPolicyTrainer
from avalon.agent.common.types import ActionBatch
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import BatchSequenceData
from avalon.agent.common.types import ObservationBatch
from avalon.agent.godot.godot_eval import test
from avalon.agent.godot.godot_gym import GodotEnvironmentParams
from avalon.contest.agents.contest_algorithm import load_algorithm
from avalon.contest.contest_params import ContestAlgorithmParams


class ContestAlgorithmWrapper(Algorithm):
    """Wraps a user-provided algorithm into our internal Algorithm class."""

    def __init__(self, params: Params, observation_space: gym.spaces.Dict, action_space: gym.spaces.Dict) -> None:
        super().__init__(params, observation_space, action_space)
        self.contest_algorithm = load_algorithm()

    def rollout_step(
        self,
        next_obs: ObservationBatch,
        dones: Tensor,  # shape (batch_size, )
        indices_to_run: list[bool],  # shape (batch_size, )
        exploration_mode: str,
    ) -> Tuple[ActionBatch, dict]:
        step_actions = self.contest_algorithm.rollout_step(next_obs, dones)
        return step_actions, {}

    def train_step(self, batch_data: BatchSequenceData, step: int) -> int:
        raise NotImplementedError


if __name__ == "__main__":
    fixed_worlds_path = os.getenv("FIXED_WORLDS_PATH", "/tmp/avalon_worlds/minival/")

    params = ContestAlgorithmParams(
        project="contest_agent",
        is_testing=True,
        is_training=False,
        env_params=GodotEnvironmentParams(fixed_worlds_load_from_path=Path(fixed_worlds_path)),
    )
    trainer = OnPolicyTrainer(params)

    assert trainer.params.observation_space is not None
    assert trainer.params.action_space is not None

    algorithm = ContestAlgorithmWrapper(params, trainer.params.observation_space, trainer.params.action_space)

    try:
        test(trainer.params, algorithm)
    finally:
        trainer.shutdown()



from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import cast

import attr
import numpy as np
from scipy import stats

from avalon.common.utils import only
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import MAX_FALL_DISTANCE_TO_DIE
from avalon.datagen.world_creation.constants import MIN_FALL_DISTANCE_TO_CAUSE_DAMAGE
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.entities.food import FoodTreeBase
from avalon.datagen.world_creation.entities.tools.placeholder import Placeholder
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.harmonics import EdgeConfig
from avalon.datagen.world_creation.worlds.obstacles.height_path import HeightPath
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class DescendTaskConfig(TaskConfig):
    # how much extra height (above and beyond the amount required to cause some damage) to add for difficulty 0.0 and
    # 1.0 respectively. Result will be normally distributed centered on that value proportional to difficulty, with
    # the given standard deviation
    extra_height_easy: float = 0.0
    extra_height_hard: float = MAX_FALL_DISTANCE_TO_DIE - MIN_FALL_DISTANCE_TO_CAUSE_DAMAGE + 5.0
    extra_height_std_dev: float = 4.0
    # how wide to make the climbable path at difficulty 0.0 and 1.0 respectively
    path_width_easy: float = 10.0
    path_width_hard: float = 1.0
    # how far away the food starts. Not a particularly important parameter for this task probably
    goal_distance_easy: float = 8.0
    goal_distance_hard: float = 16.0


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class SingleTaskDescendTaskConfig(DescendTaskConfig):
    # slightly different settings for the single task case because otherwise agents learn to grab on to trees when they
    # fall. Which, to be far, is a legit strategy for decreasing your likelihood of death when falling, but it's a
    # much worse strategy than just carefully climbing down...
    # make the height even larger in the single task case
    extra_height_hard: float = (MAX_FALL_DISTANCE_TO_DIE - MIN_FALL_DISTANCE_TO_CAUSE_DAMAGE + 5.0) * 2.0
    # make the tree spawn farther away
    goal_distance_hard: float = 32.0


def generate_descend_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: Optional[DescendTaskConfig] = None,
) -> None:
    world, locations, difficulty = create_descend_obstacle(rand, difficulty, export_config, task_config=task_config)
    world, locations = world.end_height_obstacles(
        locations, is_accessible_from_water=False, is_spawn_region_climbable=False
    )
    world = add_food_tree_for_simple_task(world, locations)
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    export_world(output_path, rand, world)


def create_descend_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    task_config: Optional[DescendTaskConfig] = None,
) -> Tuple[World, WorldLocations, float]:

    if task_config is None:
        if constraint is None:
            task_config = SingleTaskDescendTaskConfig()
        else:
            task_config = DescendTaskConfig()

    is_platform_mode, difficulty = select_boolean_difficulty(difficulty, rand)
    is_everywhere_climbable, difficulty = select_boolean_difficulty(difficulty, rand)

    desired_goal_dist = scale_with_difficulty(
        difficulty, task_config.goal_distance_easy, task_config.goal_distance_hard
    )
    extra_fall_distance = normal_distrib_range(
        task_config.extra_height_easy,
        task_config.extra_height_hard,
        task_config.extra_height_std_dev,
        rand,
        difficulty,
    )

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, 0.5), rand, difficulty, export_config, constraint
    )

    height_delta = MIN_FALL_DISTANCE_TO_CAUSE_DAMAGE + extra_fall_distance
    climbable_path_width = normal_distrib_range(
        task_config.path_width_easy, task_config.path_width_hard, 0.5, rand, difficulty
    )

    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        constraint=constraint,
        gap_distance=0.0,
        height=-height_delta,
        traversal_width=2.0,
        inner_traversal_length=0.0,
        is_single_obstacle=True,
        is_inside_climbable=is_everywhere_climbable,
        is_outside_climbable=is_everywhere_climbable,
        inner_solution=HeightSolution(
            paths=(
                HeightPath(
                    width=climbable_path_width,
                    is_path_climbable=True,
                    is_height_affected=False,
                    is_path_restricted_to_land=False,
                ),
                # this is really here to prevent the task from going over water, but without needing such a huge
                # area to be restricted to land-only (as the above, actually important path with defines where
                # things can be climbed)
                HeightPath(
                    width=1.0,
                    is_path_climbable=False,
                    is_height_affected=False,
                    is_path_restricted_to_land=True,
                ),
            ),
            # this will be replaced with the food below
            inside_items=tuple([Placeholder(position=np.array([0.0, 0.0, 0.0]))]),
            # so that the food ends up away from the edge a little bit
            inside_item_radius=0.75,
            solution_point_brink_distance=1.0,
        ),
        probability_of_centering_on_spawn=1.0,
    )
    ring_config = attr.evolve(ring_config, edge=EdgeConfig(noise=0.25))
    world = world.add_height_obstacle(rand, ring_config, locations.island)

    if constraint is None:
        new_locations = locations
        # # flatten so that you dont fall off where you spawn
        map_new = world.map.copy()
        map_new.radial_flatten(to_2d_point(new_locations.spawn), 2.0 * 2.0, extra_mask=new_locations.island)
        world = attr.evolve(world, map=map_new)
    else:
        # reset the location of the spawn to be where the placeholder ended up
        # height will be reset below
        new_locations = attr.evolve(
            locations, spawn=only([x for x in world.items if isinstance(x, Placeholder)]).position
        )
    new_items = tuple([x for x in world.items if not isinstance(x, Placeholder)])
    world = attr.evolve(world, items=new_items)

    # move the food height down
    down_vector = -1.0 * UP_VECTOR
    new_locations = attr.evolve(new_locations, goal=new_locations.goal + down_vector * height_delta)

    # if this is platform mode, make a platform to fall on so it's easier
    if is_platform_mode:
        # find the cliff edges
        sq_slope = world.map.get_squared_slope()
        cliff_edges = sq_slope > ((height_delta * 0.8) ** 2)
        island_edges = world.map.get_outline(locations.island, 3)
        cliff_edges = np.logical_and(cliff_edges, np.logical_not(island_edges))
        cliff_edges = np.logical_and(cliff_edges, locations.island)
        cliff_movement_distance = scale_with_difficulty(difficulty, 1.0, 6.0)
        for i in range(2):
            nearby = world.map.get_dist_sq_to(to_2d_point(new_locations.spawn)) < cliff_movement_distance**2
            cliff_edges = np.logical_and(cliff_edges, nearby)
            if np.any(cliff_edges):
                cliff_indices = cast(Tuple[int, int], tuple(rand.choice(np.argwhere(cliff_edges))))
                cliff_point_2d = world.map.index_to_point_2d(cliff_indices)
                cliff_position = np.array(
                    [cliff_point_2d[0], new_locations.spawn[1] - height_delta / 2.0, cliff_point_2d[1]]
                )
                cliff = FoodTreeBase(position=cliff_position)
                world = world.add_item(cliff)
                break

    if constraint is None:
        # finally, have to remove trees on harder difficulties so the agent doesn't learn to grab them to break its fall
        new_foliage_density_modifier = (
            scale_with_difficulty(difficulty, 1.0, 0.0) * world.biome_config.foliage_density_modifier
        )
        new_biome_config = attr.evolve(world.biome_config, foliage_density_modifier=new_foliage_density_modifier)
        world = attr.evolve(world, biome_config=new_biome_config)

    return world, new_locations, difficulty



from typing import List
from typing import Tuple

import attr
import numpy as np
from godot_parser import Node as GDNode
from godot_parser import NodePath
from scipy.spatial.transform import Rotation

from avalon.common.utils import only
from avalon.datagen.world_creation.entities.doors.locks.door_lock import DoorLock
from avalon.datagen.world_creation.entities.doors.types import HingeSide
from avalon.datagen.world_creation.entities.doors.types import MountSlot
from avalon.datagen.world_creation.geometry import Axis
from avalon.datagen.world_creation.indoor.blocks import new_make_block_nodes
from avalon.datagen.world_creation.indoor.constants import DEFAULT_STORY_HEIGHT
from avalon.datagen.world_creation.indoor.utils import get_scaled_mesh_node
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.utils import make_transform


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class SlidingBar(DoorLock):
    size: np.ndarray = np.array([1, DEFAULT_STORY_HEIGHT, 0.1])
    rotation: np.ndarray = np.array([1, 0, 0, 0, 1, 0, 0, 0, 1])
    sliding_axis: Axis = Axis.Y
    door_face_axis: Axis = Axis.Z
    mount_slot: MountSlot = MountSlot.BOTTOM
    mount_side: HingeSide = HingeSide.LEFT
    proportion_to_unlock: float = 0.25

    @property
    def latch_size(self) -> Tuple[float, float, float]:
        bar_width, bar_height, bar_thickness = self.size
        latch_width, latch_height, latch_thickness = bar_width * 1.6, bar_width / 2, bar_thickness * 1.05
        return latch_width, latch_height, latch_thickness

    def get_node(self, scene: GodotScene) -> GDNode:
        SLIDING_BAR = "sliding_bar"
        SLIDING_BAR_BODY = "bar_body"
        SLIDING_BAR_ANCHOR = "anchor"

        anchor_collision_mesh_node = GDNode(
            "collision_shape",
            type="CollisionShape",
            properties={
                "shape": scene.add_sub_resource("SphereShape", radius=0).reference,
                "disabled": True,
            },
        )
        anchor_node = GDNode(SLIDING_BAR_ANCHOR, "StaticBody")
        anchor_node.add_child(anchor_collision_mesh_node)

        bar_width, bar_height, bar_thickness = self.size
        leeway_multiplier = 1.2
        if self.mount_slot == MountSlot.BOTTOM:
            lower_distance = 0
            upper_distance = bar_height * self.proportion_to_unlock * leeway_multiplier
        else:
            lower_distance = -bar_height * self.proportion_to_unlock * leeway_multiplier
            upper_distance = 0

        slider_joint_node = GDNode(
            "slider_joint",
            "SliderJoint",
            properties={
                "transform": make_transform(rotation=Rotation.from_euler("z", 90, degrees=True).as_matrix().flatten()),
                "nodes/node_a": NodePath(f"../{SLIDING_BAR_ANCHOR}"),
                "nodes/node_b": NodePath(f"../{SLIDING_BAR_BODY}"),
                "linear_limit/lower_distance": lower_distance,
                "linear_limit/upper_distance": upper_distance,
                "linear_limit/softness": 0.5,
            },
        )

        x_multiplier = -1 if self.mount_side == HingeSide.LEFT else 1

        latch_width, _latch_height, _latch_thickness = self.latch_size
        # Main body of the bar
        bar_body_offset = np.array([-x_multiplier * latch_width / 2, 0, bar_thickness / 2])
        bar_body_child_nodes = new_make_block_nodes(
            scene,
            bar_body_offset,
            self.size,
            make_parent=False,
            make_mesh=False,
            collision_shape_name="body_collision_shape",
        )
        bar_body_mesh_node = get_scaled_mesh_node(
            scene, "res://entities/doors/sliding_bar.tscn", self.size, bar_body_offset
        )
        bar_body_child_nodes.append(bar_body_mesh_node)

        # Knob
        y_multiplier = 1 if self.mount_slot == MountSlot.BOTTOM else -1
        knob_thickness = 0.1
        knob_size = np.array([0.1, 0.1, knob_thickness])
        knob_offset = bar_body_offset + np.array(
            [0, y_multiplier * bar_height / 2.2, bar_thickness / 2 + knob_thickness / 2]
        )
        knob_child_nodes = new_make_block_nodes(
            scene,
            knob_offset,
            knob_size,
            make_parent=False,
            make_mesh=False,
            collision_shape_name="knob_collision_shape",
        )
        knob_mesh_node = get_scaled_mesh_node(
            scene, "res://entities/doors/bar_knob.tscn", knob_size, knob_offset, mesh_name="knob_mesh"
        )
        knob_child_nodes.append(knob_mesh_node)
        bar_body_child_nodes.extend(knob_child_nodes)

        # Guiderail on the side
        rail_height = bar_height * (1 + self.proportion_to_unlock)
        rail_width = bar_width * 1.5
        rail_thickness = bar_width * 1.5
        # The rail is normally horizontal, so we define its size as if it was going to be placed horizontal
        # and then rotate it
        rail_size = np.array([rail_height, rail_width, rail_thickness])
        rail_position = np.array(
            [
                x_multiplier * rail_width / 2,
                0.5 * self.proportion_to_unlock * bar_height,
                bar_thickness / 1.8,
            ]
        )

        rail_mesh_node = get_scaled_mesh_node(scene, "res://entities/doors/rail.tscn", rail_size)
        rotation_degrees = 90
        rail_node = only(
            new_make_block_nodes(
                scene,
                rail_position,
                rail_size,
                Rotation.from_euler("z", rotation_degrees, degrees=True).as_matrix().flatten(),
                parent_name="rail",
                make_mesh=False,
                collision_shape_disabled=True,
            )
        )
        rail_node.add_child(rail_mesh_node)

        # Guiderail "hinges" (the wheels that slide with the bar up and down)
        hinge_span = (bar_height * (1 - self.proportion_to_unlock)) * 0.7  # 70% of the part of bar above the slot
        hinge_width = bar_width * 2.1
        hinge_height = bar_width
        hinge_thickness = 0.05
        offset_from_bottom = (self.proportion_to_unlock * bar_height) / 2
        # The hinge is normally vertical, so we define its size as if it was going to be placed vertically, then rotate
        hinge_size = np.array([hinge_height, hinge_width, hinge_thickness])
        for i, y_offset in enumerate([-hinge_span / 2 + offset_from_bottom, hinge_span / 2 + offset_from_bottom]):
            hinge_position = np.array([x_multiplier * rail_width / 2, y_offset, bar_thickness / 1.8])
            y_rotation = 0 if self.mount_side == HingeSide.LEFT else 180
            rail_hinge_mesh_node = get_scaled_mesh_node(
                scene, "res://entities/doors/rail_hinge.tscn", hinge_size, mesh_name="hinge_mesh"
            )
            rail_hinge_node = only(
                new_make_block_nodes(
                    scene,
                    hinge_position,
                    hinge_size,
                    Rotation.from_euler("zy", (90, y_rotation), degrees=True).as_matrix().flatten(),
                    make_mesh=False,
                    parent_name=f"rail_hinge_{i}",
                    collision_shape_disabled=True,
                )
            )
            rail_hinge_node.add_child(rail_hinge_mesh_node)
            bar_body_child_nodes.append(rail_hinge_node)

        body_script = scene.add_ext_resource("res://entities/doors/vertical_bar_body.gd", "Script")
        body_node = GDNode(
            SLIDING_BAR_BODY,
            type="RigidBody",
            properties={
                "mass": 10.0,
                "script": body_script.reference,
                "entity_id": 0,
            },
        )
        for child_node in bar_body_child_nodes:
            body_node.add_child(child_node)

        group_spatial_node = GDNode(
            f"{SLIDING_BAR}_{self.entity_id}",
            "Spatial",
            properties={
                "transform": make_transform(position=self.position, rotation=self.rotation),
                "script": scene.add_ext_resource("res://entities/doors/sliding_bar.gd", "Script").reference,
                "sliding_axis": self.sliding_axis.value,
                "proportion_to_unlock": self.proportion_to_unlock,
                "entity_id": self.entity_id,
            },
        )
        group_spatial_node.add_child(body_node)
        group_spatial_node.add_child(anchor_node)
        group_spatial_node.add_child(slider_joint_node)
        group_spatial_node.add_child(rail_node)
        return group_spatial_node

    def get_additional_door_body_nodes(
        self, scene: GodotScene, body_size: np.ndarray, body_centroid: Point3DNP = np.array([0, 0, 0])
    ) -> List[GDNode]:
        # For the sliding bar these are the two big holes on the doors that the bar falls into
        # latch_mesh_base_size = np.array([0.15, 0.09, 0.181])
        bar_width, bar_height, bar_thickness = self.size
        door_body_width, door_body_height, door_body_thickness = body_size
        # latch_width, latch_height, latch_thickness = 0.08, 0.17, 0.29
        latch_width, latch_height, latch_thickness = self.latch_size
        latch_size = np.array([latch_width, latch_height, latch_thickness])
        x_multiplier = -1 if self.mount_side == HingeSide.LEFT else 1
        y_multiplier = -1 if self.mount_slot == MountSlot.BOTTOM else 1
        bar_y_position = self.position[1]

        bar_latch_positions = [
            np.array(
                [
                    body_centroid[0] + x_multiplier * (door_body_width / 2 - latch_width / 2),
                    bar_y_position + y_multiplier * bar_height * self.proportion_to_unlock,
                    door_body_thickness / 2 + bar_thickness / 2,
                ]
            ),
        ]
        latch_nodes = []

        for i, bar_latch_position in enumerate(bar_latch_positions):
            bar_latch_node = only(
                new_make_block_nodes(
                    scene,
                    bar_latch_position,
                    latch_size,
                    make_mesh=False,
                    parent_name=f"bar_latch_{self.entity_id}_{i}",
                    collision_shape_disabled=True,
                )
            )
            bar_latch_mesh_node = get_scaled_mesh_node(scene, "res://entities/doors/bar_latch.tscn", latch_size)
            bar_latch_node.add_child(bar_latch_mesh_node)
            latch_nodes.append(bar_latch_node)
        return latch_nodes



import math
from typing import Optional

import numpy as np
from loguru import logger

from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.constants import DEFAULT_SAFETY_RADIUS
from avalon.datagen.world_creation.debug_plots import plot_value_grid
from avalon.datagen.world_creation.entities.tools.placeholder import Placeholder
from avalon.datagen.world_creation.types import HeightMode
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.obstacles.harmonics import EdgeConfig
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.obstacles.ring_obstacle import HeightObstacle
from avalon.datagen.world_creation.worlds.obstacles.ring_obstacle import RingObstacleConfig
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.utils import add_offsets
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


def make_ring(
    rand: np.random.Generator,
    difficulty: float,
    world: World,
    locations: WorldLocations,
    gap_distance: float,
    height: float,
    traversal_width: float,
    constraint: Optional[CompositionalConstraint],
    is_inside_climbable: bool = True,
    is_outside_climbable: bool = False,
    inner_solution: Optional[HeightSolution] = None,
    outer_solution: Optional[HeightSolution] = None,
    dual_solution: Optional[HeightSolution] = None,
    inner_traversal_length: float = 0.0,
    outer_traversal_length: float = 0.0,
    safety_radius: float = DEFAULT_SAFETY_RADIUS,
    traversal_noise_interpolation_multiple: float = 2.0,
    is_single_obstacle: bool = False,
    probability_of_centering_on_spawn: Optional[float] = 0.5,
    max_additional_radius_multiple: float = 2.0,
    height_mode: HeightMode = HeightMode.RELATIVE,
    extra_safety_radius: float = 0.0,
    expansion_meters: float = 0.0,
    detail_radius: float = 20.0,
    traversal_difficulty_meters: float = 10.0,
) -> RingObstacleConfig:
    is_debugging = False

    if constraint is None:
        # should only be centered on you half the time
        assert probability_of_centering_on_spawn is not None
        is_centered_on_spawn = rand.uniform() < probability_of_centering_on_spawn
        if is_centered_on_spawn:
            center_point = to_2d_point(locations.spawn)
            other_point = to_2d_point(locations.goal)
        else:
            if is_debugging:
                logger.debug("Is reversed")
            center_point = to_2d_point(locations.goal)
            other_point = to_2d_point(locations.spawn)

            if is_single_obstacle:
                height *= -1
            else:
                # swap these because this is not the canonical ordering
                is_inside_climbable, is_outside_climbable = is_outside_climbable, is_inside_climbable
                inner_solution, outer_solution = outer_solution, inner_solution
                inner_traversal_length, outer_traversal_length = outer_traversal_length, inner_traversal_length

            if inner_solution is not None:
                inner_solution = inner_solution.reverse()

            if outer_solution is not None:
                outer_solution = outer_solution.reverse()

            if dual_solution is not None:
                dual_solution = dual_solution.reverse()
    else:
        center_point = to_2d_point(constraint.center)
        other_point = to_2d_point(locations.goal)

        if constraint.is_height_inverted:
            height *= -1

    middle_point = (center_point + other_point) / 2.0

    if constraint is None:
        # add some variation for the radius of the circle
        radius_scale = rand.uniform(1, max_additional_radius_multiple)
        center_point = middle_point + (center_point - middle_point) * radius_scale
    else:
        radius_scale = 1.0

    # and find a randomized traversal point that would actually work
    spawn_radius = np.linalg.norm(center_point - to_2d_point(locations.spawn))
    food_radius = np.linalg.norm(center_point - to_2d_point(locations.goal))
    min_traversal_radius = (
        min([spawn_radius, food_radius]) + safety_radius + inner_traversal_length + extra_safety_radius  # type: ignore
    )
    max_traversal_radius = max([spawn_radius, food_radius]) - (  # type: ignore
        safety_radius + outer_traversal_length + gap_distance + extra_safety_radius
    )
    radius_diff = max_traversal_radius - min_traversal_radius
    if radius_diff < 0:
        raise ImpossibleWorldError("Too small to make the requested ring")
    if is_debugging:
        logger.debug(f"Radius diff is {radius_diff}")
        logger.debug(spawn_radius)
        logger.debug(food_radius)
        logger.debug(min_traversal_radius)
        logger.debug(max_traversal_radius)
        logger.debug(safety_radius)
        logger.debug(inner_traversal_length)
        logger.debug(radius_scale)
    min_selectable_radius = min_traversal_radius + radius_diff * 0.3
    max_selectable_radius = max_traversal_radius - radius_diff * 0.3
    r_squared = world.map.get_dist_sq_to(center_point)
    acceptable_radius_mask = np.logical_and(
        r_squared < max_selectable_radius**2, r_squared > min_selectable_radius**2
    )
    if is_debugging:
        plot_value_grid(r_squared < max_selectable_radius**2, "Less than max radius")
        plot_value_grid(r_squared > min_selectable_radius**2, "Greater than min radius")
        plot_value_grid(acceptable_radius_mask, "Acceptable radius mask")

    max_dist = 2.0 + traversal_difficulty_meters * difficulty

    temp = middle_point - center_point
    middle_theta = np.arctan2(temp[1], temp[0])
    r = (min_traversal_radius + max_traversal_radius) / 2.0
    actual_middle = center_point + np.array([r * math.cos(middle_theta), r * math.sin(middle_theta)])

    # logger.debug(middle_theta)
    # logger.debug(middle_point)
    # logger.debug(center_point)
    # logger.debug(r)
    # logger.debug(actual_middle)

    if constraint is None:
        dist_sq_to_middle_point = world.map.get_dist_sq_to(actual_middle)
        near_traversal_center_mask = dist_sq_to_middle_point < max_dist**2
    else:
        near_traversal_center_mask = constraint.traversal_mask
    possible_points = np.logical_and(
        locations.island, np.logical_and(near_traversal_center_mask, acceptable_radius_mask)
    )
    if not np.any(possible_points):
        traversal_indices_array = None
    else:
        # biasing away from the ocean so we end up with failing paths less often
        if constraint is None:
            # TODO: max points happens to be reduced under us, so annoying :(
            water_dist = world.map.get_water_distance(
                rand, is_fresh_water_included_in_moisture=True, max_points=600, for_points=possible_points
            )
            if water_dist is not None:
                water_dist_for_points = water_dist[possible_points]
                power_to_raise = 8
                weights = (water_dist_for_points / water_dist_for_points.max()) ** power_to_raise
                weights /= weights.sum()
            else:
                weights = None
            traversal_indices_array = rand.choice(np.argwhere(possible_points), p=weights)

            # pd.DataFrame(water_dist[water_dist > 0.0]).hist()
            # plot_value_grid(water_dist, "water dist", markers=[tuple(traversal_indices_array)])
            # plot_value_grid(
            #     (water_dist / water_dist.max()) ** power_to_raise,
            #     "probabilities",
            #     markers=[tuple(traversal_indices_array)],
            # )
            # plot_value_grid(world.map.Z)
        else:
            traversal_indices_array = rand.choice(np.argwhere(possible_points))
    if is_debugging:
        plot_value_grid(near_traversal_center_mask, "Near traversal center")
        plot_value_grid(possible_points, "Possible points")

    if traversal_indices_array is None:
        raise ImpossibleWorldError("Could not find valid traversal point")
        # fine, whatever, just set to the midpoint
        # ring_traversal_point = to_2d_point((locations.spawn + locations.goal) / 2.0)
    else:
        traversal_indices = tuple(traversal_indices_array)
        ring_traversal_point = np.array([world.map.X[traversal_indices], world.map.Y[traversal_indices]])

    if is_debugging:
        logger.debug(f"Selected traversal point {ring_traversal_point}")
        logger.debug(np.linalg.norm(ring_traversal_point - center_point))

    if is_single_obstacle:
        outer_obstacle = None
        assert outer_solution is None, "Well that doesnt make sense"
    else:
        outer_obstacle = HeightObstacle(
            edge_config=EdgeConfig(),
            is_inside_ring=False,
            traversal_length=outer_traversal_length,
            traversal_width=traversal_width,
            traversal_noise_interpolation_multiple=traversal_noise_interpolation_multiple,
            is_default_climbable=is_outside_climbable,
            detail_radius=detail_radius,
        )

    ring_config = RingObstacleConfig(
        center_point=center_point,
        traversal_point=ring_traversal_point,
        edge=EdgeConfig(),
        height=height,
        height_mode=height_mode,
        inner_safety_radius=min([spawn_radius, food_radius]) + (safety_radius + extra_safety_radius),  # type: ignore
        outer_safety_radius=max([spawn_radius, food_radius]) - (safety_radius + extra_safety_radius),  # type: ignore
        inner_obstacle=HeightObstacle(
            edge_config=EdgeConfig(),
            is_inside_ring=True,
            traversal_length=inner_traversal_length,
            traversal_width=traversal_width,
            traversal_noise_interpolation_multiple=traversal_noise_interpolation_multiple,
            is_default_climbable=is_inside_climbable,
            detail_radius=detail_radius,
        ),
        outer_obstacle=outer_obstacle,
        chasm_bottom_size=gap_distance,
        inner_solution=inner_solution,
        outer_solution=outer_solution,
        dual_solution=dual_solution,
        expansion_meters=expansion_meters,
    )

    return ring_config


def create_outer_placeholder_solution(
    count: int = 5, offset: float = 0.5, randomization_dist: float = 2.0
) -> HeightSolution:
    items = [Placeholder(offset=offset, position=np.array([-0.5 * i, 0.0, 0.0])) for i in range(count)]
    return HeightSolution(
        outside_items=tuple(add_offsets(items)),
        outside_item_randomization_distance=randomization_dist,
        outside_item_radius=items[0].get_offset(),
        solution_point_brink_distance=0.5,
    )


def create_inner_placeholder_solution(
    count: int = 5, offset: float = 0.5, randomization_dist: float = 2.0
) -> HeightSolution:
    items = [Placeholder(offset=offset, position=np.array([-0.5 * i, 0.0, 0.0])) for i in range(count)]
    return HeightSolution(
        inside_items=tuple(add_offsets(items)),
        inside_item_randomization_distance=randomization_dist,
        inside_item_radius=items[0].get_offset(),
        solution_point_brink_distance=0.5,
    )



import math
from enum import Enum
from typing import Dict
from typing import Final
from typing import List
from typing import Sequence
from typing import Set
from typing import Tuple

import numpy as np


class AvalonTask(Enum):
    # ONE
    EAT = "EAT"
    MOVE = "MOVE"
    # TWO
    JUMP = "JUMP"
    EXPLORE = "EXPLORE"
    SCRAMBLE = "SCRAMBLE"
    # THREE
    CLIMB = "CLIMB"
    DESCEND = "DESCEND"
    THROW = "THROW"
    AVOID = "AVOID"
    # FOUR
    HUNT = "HUNT"
    FIGHT = "FIGHT"
    PUSH = "PUSH"
    STACK = "STACK"
    BRIDGE = "BRIDGE"
    OPEN = "OPEN"
    CARRY = "CARRY"
    # COMPOSITIONAL
    NAVIGATE = "NAVIGATE"
    FIND = "FIND"
    GATHER = "GATHER"
    SURVIVE = "SURVIVE"


avalon_task_to_int = {member.value: i for i, member in enumerate(AvalonTask)}
int_to_avalon_task = {i: member.value for i, member in enumerate(AvalonTask)}


class AvalonTaskGroup(Enum):
    ONE = "ONE"
    TWO = "TWO"
    THREE = "THREE"
    FOUR = "FOUR"
    SIMPLE = "SIMPLE"
    EASY = "EASY"
    COMPOSITIONAL = "COMPOSITIONAL"
    ALL = "ALL"
    # ONE
    EAT = "EAT"
    MOVE = "MOVE"
    # TWO
    JUMP = "JUMP"
    EXPLORE = "EXPLORE"
    SCRAMBLE = "SCRAMBLE"
    # THREE
    CLIMB = "CLIMB"
    DESCEND = "DESCEND"
    THROW = "THROW"
    AVOID = "AVOID"
    # FOUR
    HUNT = "HUNT"
    FIGHT = "FIGHT"
    PUSH = "PUSH"
    STACK = "STACK"
    BRIDGE = "BRIDGE"
    OPEN = "OPEN"
    CARRY = "CARRY"
    # COMPOSITIONAL
    NAVIGATE = "NAVIGATE"
    FIND = "FIND"
    GATHER = "GATHER"
    SURVIVE = "SURVIVE"


TRAIN_TASK_GROUPS = (
    AvalonTaskGroup.ONE,
    AvalonTaskGroup.TWO,
    AvalonTaskGroup.THREE,
    AvalonTaskGroup.FOUR,
    AvalonTaskGroup.COMPOSITIONAL,
)

SINGLE_TASK_GROUPS = (
    AvalonTaskGroup.EAT,
    AvalonTaskGroup.MOVE,
    AvalonTaskGroup.JUMP,
    AvalonTaskGroup.CLIMB,
    AvalonTaskGroup.DESCEND,
    AvalonTaskGroup.SCRAMBLE,
    AvalonTaskGroup.STACK,
    AvalonTaskGroup.BRIDGE,
    AvalonTaskGroup.PUSH,
    AvalonTaskGroup.THROW,
    AvalonTaskGroup.HUNT,
    AvalonTaskGroup.FIGHT,
    AvalonTaskGroup.AVOID,
    AvalonTaskGroup.EXPLORE,
    AvalonTaskGroup.OPEN,
    AvalonTaskGroup.CARRY,
    AvalonTaskGroup.NAVIGATE,
    AvalonTaskGroup.FIND,
    AvalonTaskGroup.SURVIVE,
    AvalonTaskGroup.GATHER,
)

TASKS_BY_TASK_GROUP: Dict[AvalonTaskGroup, Tuple[AvalonTask, ...]] = {
    AvalonTaskGroup.ONE: (AvalonTask.EAT, AvalonTask.MOVE),
    AvalonTaskGroup.TWO: (AvalonTask.JUMP, AvalonTask.EXPLORE, AvalonTask.SCRAMBLE),
    AvalonTaskGroup.THREE: (AvalonTask.CLIMB, AvalonTask.DESCEND, AvalonTask.THROW, AvalonTask.AVOID),
    AvalonTaskGroup.FOUR: (
        AvalonTask.HUNT,
        AvalonTask.FIGHT,
        AvalonTask.PUSH,
        AvalonTask.STACK,
        AvalonTask.BRIDGE,
        AvalonTask.OPEN,
        AvalonTask.CARRY,
    ),
    AvalonTaskGroup.COMPOSITIONAL: (AvalonTask.NAVIGATE, AvalonTask.FIND, AvalonTask.GATHER, AvalonTask.SURVIVE),
    AvalonTaskGroup.EASY: (
        AvalonTask.EAT,
        AvalonTask.MOVE,
    ),
    AvalonTaskGroup.SIMPLE: (
        AvalonTask.EAT,
        AvalonTask.MOVE,
        AvalonTask.JUMP,
        AvalonTask.CLIMB,
        AvalonTask.DESCEND,
        AvalonTask.SCRAMBLE,
        AvalonTask.STACK,
        AvalonTask.BRIDGE,
        AvalonTask.PUSH,
        AvalonTask.THROW,
        AvalonTask.HUNT,
        AvalonTask.FIGHT,
        AvalonTask.AVOID,
        AvalonTask.EXPLORE,
        AvalonTask.OPEN,
        AvalonTask.CARRY,
    ),
    AvalonTaskGroup.ALL: (
        AvalonTask.EAT,
        AvalonTask.MOVE,
        AvalonTask.JUMP,
        AvalonTask.CLIMB,
        AvalonTask.DESCEND,
        AvalonTask.SCRAMBLE,
        AvalonTask.STACK,
        AvalonTask.BRIDGE,
        AvalonTask.PUSH,
        AvalonTask.THROW,
        AvalonTask.HUNT,
        AvalonTask.FIGHT,
        AvalonTask.AVOID,
        AvalonTask.EXPLORE,
        AvalonTask.OPEN,
        AvalonTask.CARRY,
        AvalonTask.NAVIGATE,
        AvalonTask.FIND,
        AvalonTask.SURVIVE,
        AvalonTask.GATHER,
    ),
    AvalonTaskGroup.EAT: (AvalonTask.EAT,),
    AvalonTaskGroup.MOVE: (AvalonTask.MOVE,),
    AvalonTaskGroup.JUMP: (AvalonTask.JUMP,),
    AvalonTaskGroup.CLIMB: (AvalonTask.CLIMB,),
    AvalonTaskGroup.DESCEND: (AvalonTask.DESCEND,),
    AvalonTaskGroup.SCRAMBLE: (AvalonTask.SCRAMBLE,),
    AvalonTaskGroup.STACK: (AvalonTask.STACK,),
    AvalonTaskGroup.BRIDGE: (AvalonTask.BRIDGE,),
    AvalonTaskGroup.PUSH: (AvalonTask.PUSH,),
    AvalonTaskGroup.THROW: (AvalonTask.THROW,),
    AvalonTaskGroup.HUNT: (AvalonTask.HUNT,),
    AvalonTaskGroup.FIGHT: (AvalonTask.FIGHT,),
    AvalonTaskGroup.AVOID: (AvalonTask.AVOID,),
    AvalonTaskGroup.EXPLORE: (AvalonTask.EXPLORE,),
    AvalonTaskGroup.OPEN: (AvalonTask.OPEN,),
    AvalonTaskGroup.CARRY: (AvalonTask.CARRY,),
    AvalonTaskGroup.NAVIGATE: (AvalonTask.NAVIGATE,),
    AvalonTaskGroup.FIND: (AvalonTask.FIND,),
    AvalonTaskGroup.SURVIVE: (AvalonTask.SURVIVE,),
    AvalonTaskGroup.GATHER: (AvalonTask.GATHER,),
}

AGENT_HEIGHT = 2.0
HALF_AGENT_HEIGHT_VECTOR = np.array([0, AGENT_HEIGHT / 2.0, 0])

FOOD_HOVER_DIST = 0.5

MAX_FLAT_JUMP_METERS = 3
MAX_JUMP_HEIGHT_METERS = 1.6
MAX_MOVE_DIST = 10
MAX_REACH_DIST = 0.3
STANDING_REACH_HEIGHT = AGENT_HEIGHT + MAX_REACH_DIST

# minimum height that prevents you from escaping without climbinb
JUMPING_REQUIRED_HEIGHT = 2.0
CLIMBING_REQUIRED_HEIGHT = 3.2

ITEM_FLATTEN_RADIUS = 2.0

# how big is the stone for stacking
BOX_HEIGHT = 1.0

# how big is the boulder for pushing
BOULDER_HEIGHT = 1.4
BOULDER_MIN_MASS = 600.0
BOULDER_MAX_MASS = 2400.0

# controls how much space to reserve around the spawn and goal points
DEFAULT_SAFETY_RADIUS = 0.5

# how far you can jump in a totally flat world
MAX_FLAT_JUMP_DIST = 2.3
# how far you can effectively jump given polygonization
MAX_EFFECTIVE_JUMP_DIST = 2.1

MIN_FALL_DISTANCE_TO_CAUSE_DAMAGE = 10.0
MAX_FALL_DISTANCE_TO_DIE = 30.0

# min dist that requires a bridge
MIN_BRIDGE_DIST = MAX_FLAT_JUMP_DIST + 0.25

# length of the bridge object
BRIDGE_LENGTH = 6.4
MAX_BRIDGE_DIST = BRIDGE_LENGTH - 2

# how far off the ground to spawn weapons. Not really sure what this should be, needs to work for stick and rock...
WEAPON_HEIGHT_OFFSET = 1.0

# standard deviation for when we're trying to tightly specify a distance for a goal
TIGHT_DIST_STD_DEV = 0.5


# any tree within this many meters of your line of sight will be deleted
# (only applies to tasks that require being able to see the goal)
METERS_OF_TREE_CLEARANCE_AROUND_LINE_OF_SIGHT = 5.0

# do not spawn ground cover within this radius of important points
FLORA_REMOVAL_METERS = 1.0


def get_all_tasks_for_task_groups(task_groups: Sequence[AvalonTaskGroup]) -> List[AvalonTask]:
    result: Set[AvalonTask] = set()
    for task_group in task_groups:
        result.update(TASKS_BY_TASK_GROUP[task_group])
    return sorted(result, key=lambda x: x.value)


STARTING_HIT_POINTS: float = 1.0

# for performance experiments
IS_ALL_SCENERY_IN_MAIN = False
FLORA_RADIUS_IN_TILES = 1

IS_DEBUGGING_IMPOSSIBLE_WORLDS = False
WATER_LINE = 0.0
UNCLIMBABLE_BIOME_ID = 19
DARK_SHORE = UNCLIMBABLE_BIOME_ID + 1
FRESH_WATER = DARK_SHORE + 1
SWAMP = FRESH_WATER + 1
MIN_CLIFF_TERRAIN_ANGLE = (math.pi / 2) * 0.4
MAX_CLIFF_TERRAIN_ANGLE = (math.pi / 2) * 0.7
WORLD_RAISE_AMOUNT = 10_000
IDENTITY_BASIS: Final = np.array([1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0], dtype=float)
UP_VECTOR = np.array([0.0, 1.0, 0.0])



import attr

from avalon.agent.common.params import ClippedNormalMode
from avalon.agent.common.params import Params
from avalon.contrib.utils import FILESYSTEM_ROOT


@attr.s(auto_attribs=True, frozen=True)
class OffPolicyParams(Params):
    worker_managers: int = 1
    # ratio between these two affects sample efficiency
    num_steps: int = 500  # env steps (per worker) per loop. this is how often we reload a new model and log stats.
    train_steps: int = 100  # train steps per training iteration
    min_fragment_len: int = 50  # min length of an episode fragment to train on
    max_fragment_len: int = 50  # max length of an episode fragment to train on

    data_dir: str = f"{FILESYSTEM_ROOT}/data/rollouts/"
    num_dataloader_workers: int = 1
    prioritize_ends: bool = True  # in dreamerv2, this is true by default and false in dmc_vision and dmc_proprio
    separate_ongoing: bool = True  # don't train on incomplete episodes
    prefill_steps: int = 5000  # only used in the danijar-replica trainer
    pretrain_steps: int = 0
    replay_buffer_size_timesteps: int = 1_000_000
    observation_compression: bool = False

    rollout_model_update_freq: int = 100  # how many training steps between pushing a new model

    log_rollout_metrics_every: int = 20  # how many rollout episodes between logging stats
    prefill_eps_per_dataloader: int = 5  # how many episodes to load before a dataloader starts serving batches
    multiprocessing_mode: str = "spawn"  # currently, spawn is necessary for the off policy to work. could be fixed.

    @property
    def replay_buffer_size_timesteps_per_manager(self) -> int:
        return int(self.replay_buffer_size_timesteps / self.worker_managers)


@attr.s(auto_attribs=True, frozen=True)
class DreamerParams(OffPolicyParams):
    """Note: these are all the "default" params in the dreamerv2 config.
    PLEASE DON'T CHANGE THESE!
    I want to maintain this config for replicating danijar's repo.
    """

    # overrides
    project: str = "avalon_dreamer"
    clip_grad_norm: float = 100.0
    reward_loss_scale: float = 1
    discount: float = 0.99
    obs_first: bool = False
    normal_std_from_model: bool = True
    clipped_normal_mode: ClippedNormalMode = ClippedNormalMode.TRUNCATED_NORMAL
    policy_normal_init_std: float = (
        0  # Seems strange to start with 0 (well, .1) std but afaict this is how dreamerv2 does it
    )
    policy_normal_min_std: float = 0.1
    replay_buffer_size_timesteps: int = 2_000_000
    min_fragment_len: int = 50
    max_fragment_len: int = 50
    prioritize_ends: bool = True
    batch_size: int = 16

    # new dreamer stuff
    horizon: int = 15  # how long to roll out imagination
    stoch_size: int = 32  # the size of the stochastic part of the latent. only the output shape of some MLPs
    # the size of the deterministic part of the latent. also the hidden and state size of the GRU
    # this defaulted to 200 in dreamerv1, and 1024 in dreamerv2 (but 200 for dmc, 600 atari)
    deter_size: int = 1024
    # hidden layer size for MLPs in RSSM model
    # note that this is 1024 default in dreamerv2 (but 200 in dmc, 600 in atari), and 400 in dreamerv1
    rssm_hidden_size: int = 1024
    encoder_mlp_hidden_size: int = 400
    # 400 in all dreamerv2 configs
    decoder_mlp_hidden_size: int = 400

    model_lr: float = 1e-4
    value_lr: float = 2e-4
    actor_lr: float = 8e-5
    adam_eps: float = 1e-5
    adam_weight_decay: float = 1e-6

    free_nats: float = 0.0
    kl_loss_scale: float = 1.0
    kl_balance: float = 0.8  # trains posterior faster than prior
    disclam: float = 0.95  # lambda for GAE
    # 1e-4 is the value used in dreamer for walker walk. Much larger in atari.
    policy_entropy_scale: float = 2e-3
    pcont: bool = True  # predict done signals
    pcont_loss_scale: float = 1.0  # weight of pcont loss
    actor_grad: str = "dynamics"  # which type of learning to use for the actor, "dynamics" to backprop through model
    freeze_actor_steps: int = 0  # don't train the actor for the first n steps to let the world model stabilize a bit
    value_target_network_update_freq: int = 100

    # TODO: figure out why "eval" doesn't work well, when that's what danijar always used.
    eval_exploration_mode: str = "explore"  # "eval" to use the mode of the distribution, "explore" to sample
    eval_workers: int = 16
    train_every: int = 5  # only used in the dreamerv2-style trainer. how many env steps per training step.

    def __attrs_post_init__(self) -> None:
        # These should only be changed with great care.
        assert self.obs_first is False
        assert self.time_limit_bootstrapping is False



from typing import Final

import attr
from godot_parser import GDObject
from godot_parser import Node as GDNode
from scipy.spatial.transform import Rotation

from avalon.datagen.world_creation.entities.item import Item
from avalon.datagen.world_creation.types import GodotScene

PLAYER_SPAWN_POINT: Final = "SpawnPoint"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class SpawnPoint(Item):

    pitch: float
    yaw: float
    is_dynamic: bool = False
    is_visibility_required: bool = True
    entity_id: int = -1

    name: str = PLAYER_SPAWN_POINT

    def get_node(self, scene: GodotScene) -> GDNode:
        spawn_yaw_rotation = Rotation.from_euler("y", self.yaw, degrees=True)
        spawn_pitch_rotation = Rotation.from_euler("x", self.pitch, degrees=True)
        spawn_rotation = (spawn_yaw_rotation * spawn_pitch_rotation).as_matrix().flatten()
        spawn_transform = GDObject("Transform", *spawn_rotation, *self.position)
        return GDNode(self.name, type="Spatial", properties={"transform": spawn_transform})



import torch
from torch import Tensor

from avalon.agent.common.types import ActionBatch
from avalon.agent.common.types import ObservationBatch


class RandomAlgorithm:
    """
    Instruction to participant: replace this class with your own algorithm.
    """

    def __init__(self) -> None:
        torch.manual_seed(0)
        self.real_dist = torch.distributions.Normal(loc=torch.tensor([0.0] * 18), scale=torch.tensor([1.0] * 18))
        self.discrete_dist = torch.distributions.Bernoulli(probs=torch.tensor([0.5] * 3))

    def rollout_step(
        self,
        next_obs: ObservationBatch,
        dones: Tensor,  # shape (batch_size, )
    ) -> ActionBatch:
        batch_size = next_obs["rgbd"].shape[0]
        real_actions = self.real_dist.sample((batch_size,))

        discrete_actions = torch.zeros((batch_size, 3, 2))
        discrete_sample = self.discrete_dist.sample((batch_size,))
        discrete_actions[:, :, 1] = discrete_sample
        discrete_actions[:, :, 0] = 1 - discrete_sample

        step_actions: ActionBatch = {"real": real_actions, "discrete": discrete_actions}

        return step_actions


def load_algorithm():
    return RandomAlgorithm()



import attr

from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class CompositionalConstraint:
    locations: WorldLocations
    world: World
    center: Point3DNP
    traversal_mask: MapBoolNP
    is_height_inverted: bool
